/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_capture.h
 * purposes:	definitions and data structures for the main capture thread
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef __SELFDHCP_CAPTURE_H
#define __SELFDHCP_CAPTURE_H

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pcap.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <net/if.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <signal.h>

#include "selfdhcp_struct.h"
#include "selfdhcp_pthread.h"
#include "selfdhcp_profiles.h"
#include "selfdhcp_arpp.h"
#include "selfdhcp_heuristics.h"

/* extern variables passed from the main function */
extern int program_flags;
extern char *prog_name;

/* Default interface to be used if none is specified */
#ifdef __SOLARIS
#define DEFAULT_IFACE	"hme0"
#else
#define DEFAULT_IFACE	"eth0"
#endif

/* pcap filter: we only want to capture packets which are useful, i.e. contain
 * MAC & IP couples */
#define BPF_FILTER_STR	"arp or ip and not ip multicast"

/* maximum dimension for the packets to be captured for pcap_open_live() */
#define MAX_BYTE 	1500
/* timeout to read a packet for pcap_open_live() */
#define T_OUT 		100

/* dimension for the circular buffer */
#define CIRC_BUFF_DIM		200

/* structure to hold an IP-MAC couple for the circular buffer */
typedef struct couple
{
  struct in_addr ip_addr;
  u_char mac_addr[ETH_ALEN];
} couple_t;

/* this structure is used to pass arguments to the capture_start() and other
 * functions, because pthreads requires to pass a single **void argument */
typedef struct capture_args
{
  char *device;
  lista_t **liste;
  netconfig_t *net_conf;
  ip2test_t *iplist;
} capture_args_t;

/* function capture_start()
 * this function is used to start the packet capture sequence. Both the name of
 * the interface to be used and the pointer to the list where MAC and IP
 * addresses are to be stored is bassed through the void** array.
 * libpcap is used both for opening the NIC and for sniffing packets; please 
 * refer to the pcap documentetion for further info (www.tcpdump.org).
 * capture_start() returns -1 if errors are encountered, 0 else.
 */
void capture_start (void **void_args);

/* these defines are for the port used for some common services that are
 * in some way useful to selfdhcp */
#define DNSPORT			53
/* DHCP has defined server and client port */
#define DHCPPORT_SRV	67
#define DHCPPORT_CLI	68
#define SMBPORT			138

/* function packet_process()
 * this is a standard callback function for pcap_loop(), as specified in the 
 * documentation. This function should be not used directly.
 */
void packet_process (u_char * args, const struct pcap_pkthdr *header,
		     const u_char * packet);

/* we have to redefine this, because both the linux includes and libnet from
 * version 1.1 don't define any more the addressed, because of damn ipv6 I think
 */
struct arp_hdr
{
  unsigned short int ar_hrd;	/* Format of hardware address.  */
  unsigned short int ar_pro;	/* Format of protocol address.  */
  unsigned char ar_hln;		/* Length of hardware address.  */
  unsigned char ar_pln;		/* Length of protocol address.  */
  unsigned short int ar_op;	/* ARP opcode (command).  */
  unsigned char ar_sha[ETH_ALEN];	/* Sender hardware address.  */
  unsigned char ar_spa[4];	/* Sender IP address.  */
  unsigned char ar_tha[ETH_ALEN];	/* Target hardware address.  */
  unsigned char ar_tpa[4];	/* Target IP address.  */
};

/* for DHCP info refer to rfc951, rfc1497, rfc1533, rfc2131 */
/* internal structure of a DHCP or BOOTP packet, as defined in rfc951 */
typedef struct bootp_pack
{
  uint8_t op;
  uint8_t htype;
  uint8_t hlen;
  uint8_t hops;
  uint32_t xid;
  uint16_t secs;
  uint16_t flags;
  uint32_t ciaddr;
  uint32_t yiaddr;
  uint32_t siaddr;
  uint32_t giaddr;
  uint8_t chaddr[16];
  char sname[64];
  char file[128];
  uint8_t *options;		/* this field is of variable lenght, up to 512 bytes.
				 * Because it's organized in octets, uint8_t type is 
				 * used. */
} bootp_pack_t;

#define BOOTREPLY	2

/* function handle_dns()
 * handle_dns() is used to set the dns config (which in theory should be a
 * business of the heuristic thread) within the capture thread.
 */
void handle_dns (struct in_addr *ipaddr, netconfig_t * net_conf);

/* function handle_dhcp()
 * given an IP packet containing DHCP or BOOTP informations, handle_dhcp()
 * extracts useful hints such as the netmask, the net address, the gateway IP,
 * etc. Informations are stored directly in net_conf.
 */
void handle_dhcp (struct ip *ip_pack, netconfig_t * net_conf);

/* function handle_smb()
 * This function handles the insertion of the informations taken from a smb
 * broadcast packet in the net_conf structure. Right for now it simply calls
 * netconf_modify(). */
void handle_smb (uint32_t net_bcast, netconfig_t * net_conf);

#endif
