/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_heuristics.h
 * purposes:	definitions and data structures for the heuristic thread
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef __SELFDHCP_HEURISTICS_H
#define __SELFDHCP_HEURISTICS_H

#include <libnet.h>
#include <net/if_arp.h>
#include <pthread.h>
#include <unistd.h>
#include <pcap.h>
#include <netdb.h>

#include "selfdhcp_errlib.h"
#include "selfdhcp_profiles.h"
#include "selfdhcp_struct.h"
#include "selfdhcp_pthread.h"
#include "selfdhcp_arpp.h"

/* the default timeout (in minutes) for the capture process to get enough
 * packets for the heuristic thread to work */
#define DEFAULT_CAPTURE_TIMEOUT		5

/* the granularity with which the above timeout is checked (in seconds), or in
 * other words the amount of time between one try with heuristic and the next.
 * Dividing the above time for TIME_GRANULARITY, you can get the number of
 * times the heuristics are run. This parameter influences the "sensibility"
 * of selfDHCP, and should be set to greater values if the network has very
 * low traffic, or raised if the traffic is high and you think selfDHCP takes
 * too much time finding the configuration. */
#define TIME_GRANULARITY			30

/* extern variables passed from the main function */
extern int program_flags;
extern char *prog_name;

/* used for error handling in try_config(); see 'man gethostbyname' */
extern int h_errno;

/* this struture is used to hold data to be passe to any packet injection
 * function - as the one that send ARP request to test against the
 * correctness of the gateway's IP, orthe one that search for a free IP.
 */
typedef struct pack_info
{
  struct in_addr *src_ip, *dst_ip;
  u_char *src_mac, *dst_mac;
  char *device;			/* device name */
#ifndef NEW_LIBNET_INTERFACE
  u_char *packetbuff;		/* packet buffer allocated by libnet */
  struct libnet_link_int *net_link;	/* handle for the link layer */
#else
  libnet_t *lnet_struct;	/* version 1.1 uses a different structure */
#endif
} pack_info_t;

typedef struct test_hosts
{
  int num;
  uint32_t *test_IPs;
} test_hosts_t;

/* function heuristics_start()
 * this function is used to start the heuristic process. 
 */
int heuristics_start (char *device, lista_t ** liste, int timeout,
		      ip2test_t * iplist, netconfig_t * net_conf,
		      test_hosts_t * hosts);

/* function find_gatewayMAC()
 * find_gateway() scans the list of MAC addresses to find out if there is one
 * that could be the gateway. In that case, it updates tha net_conf structure
 * and returns 1,else returns 0.
 */
int find_gatewayMAC (netconfig_t * net_conf, lista_t ** liste);

/* function find_gatewayIP()
 * function find_gatewayIP() is used, once we have the gateway's MAC and the
 * netmask, to find what the gateway's IP could be.
 * If one can be found, the net_conf entry is updated and 1 is returned; 
 * else 0 is returned.
 */
int find_gatewayIP (netconfig_t * net_conf, lista_t ** liste,
		    pack_info_t * pack, ip2test_t * iplist);

/* function find_netmask()
 * this function do the heuristic work on the IP list, to find out what the
 * netmask can be. The net_conf struct must already contain a valid MAC entry
 * for the gateway, and is updated with the new netmask. 
 */
void find_netmask (netconfig_t * net_conf, lista_t ** liste);

/* function find_myip()
 * find_myip() scans through the IP list, searching for a not yet in use IP
 * address. Of course the address must not only be unused, but also in the
 * netmask. find_myip() updates the current net_conf structure, which must
 * already contain valid gateway and netmask information, with the IP
 * address it found, and returns 1; on failure, 0 is returned and the net_conf
 * structure is not modified.
 */
int find_myip (netconfig_t * net_conf, lista_t ** liste, pack_info_t * pack);

/* function is_not_relative()
 * is_not_relative() tells wether a given IP called ip2check is within relatives
 * of the gateway or not. If ip2check wasn't found within the gateway's
 * relatives, 1 is returned, if it was or in case of error, is_not_relative()
 * returns 0.
 */
int is_not_relative (u_char * gwMAC, struct in_addr *ip2check,
		     lista_t ** liste);

/* function netmask_adjust()
 * this function is used to properly adjust the netmask afer the heuristic
 * calculation (this is because the netmask must be a serie of 1 followed by 0,
 * but they must not mix. netmask_adjust() finds the first 0 in the calculated
 * netmask, and then zeroes all the following bits.
 */
void netmask_adjust (struct in_addr *netmask);

/* function is_in_netmask()
 * this function tells wether a given IP address ip2check is in the same netmask
 * with the gateway passed through the net_conf (returns 1) or not (returns 0). 
 */
int is_in_netmask (netconfig_t * net_conf, struct in_addr *ip2check);

/* function order_dns()
 * order_dns() determines which between the dns found should be the first and
 * which the second.
 */
void order_dns (netconfig_t * net_conf, lista_t ** liste);

/* time (in useconds) to wait a reply (ARP or ICMP or other) */
#define REPLY_WAIT_TIMEOUT			100000

/* time (in useconds) to wait between two ARP or ICMP request */
#define TIME_BETWEEN_SEND			5000

/* number of ARP requests to send out at once */
#define ARP_REQ_2SEND				3

/* this structure holds arguments to be passed to capture subthreads as void* */
typedef struct cap_args
{
  pcap_t *cap_session;
  int replies;
} cap_args_t;

/* function arp_req()
 * arp_req() is used to create and inject an ARP reply. Informations about the
 * addresses and everything are passed through the pack_info structure. If some
 * addresses are found to be NULL, they're filled with 0.
 * The replies are captured and filtered, and net_conf is automatically updated.
 * arp_req() return the number of replies received.
 */
int arp_req (pack_info_t * packet_info);

/* function test_config()
 * test_config() is used to create and inject some fake ICMP ECHO requests.
 * 1 is returned if at least one of these request got reply, 0 if nobody
 * answered.
 */
int test_config (netconfig_t * net_conf, pack_info_t * packet_info,
		 test_hosts_t * hosts, arpp_args_t * argums);

/* this array contains the IPs of some well known and always online hosts, to
 * be used in the try_config() function for testing. Four of the root DNS
 * servers have been chosen:
 * 202.12.27.33, 193.0.14.129, 192.36.148.17, 192.5.5.241
 */
#define DEFAULT_TEST_IP 		0xCA0C1B21, 0xC1000E81, 0xC0249411, 0xC00505F1
#define DEFAULT_TEST_IP_NUM		4

/* ICMP ECHO requests need a pad. This is the one produced by Win98 hosts ;-) */
#define ICMP_PAD	"abcdefghijklmnopqrstuvwabcdefghi"

/* function active_search()
 * this function actively sends ARP request to the gateway of every profile
 * loaded. If one gives reply, the search is stopped and the profile is copied
 * into the current network configuration. 1 is returned if a profiles fitted,
 * 0 else.
 */
int active_search (netconfig_t * net_conf, pack_info_t * pack);

#endif
