/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_profiles.h
 * purposes:	definitions and data structures for the profiles and network
 *				configurations management
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef __SELFDHCP_PROFILES_H
#define __SELFDHCP_PROFILES_H

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#ifdef __SOLARIS
#include <sys/sockio.h>
#endif
#include <netinet/in.h>
#include <net/if.h>
#include <net/route.h>
#include <libxml/xmlmemory.h>
#include <libxml/parser.h>

#include "selfdhcp_errlib.h"
#include "selfdhcp_struct.h"

/* default file in which profiles and configuration are stored */
#ifndef DEFAULT_PROFILE_FILE
#error "Run configure before make"
#endif
#ifndef DEFAULT_CONFIG_FILE
#error "Run configure before make"
#endif

/* extern variables passed from the main function */
extern int program_flags;
extern char *prog_name;

/* used for error handling; have a look at 'man errno' */
extern int errno;

/* this struct contains previously tried configurations. Used for the heuristic
 * part of selfDHCP, and for profiles management */
typedef struct _netconfig
{
  unsigned char gatewayMAC[ETH_ALEN];
  struct in_addr gatewayIP;
  struct in_addr dns1, dns2;
  struct in_addr netmask, netaddr, netbcast;
  unsigned char myMAC[ETH_ALEN];
  struct in_addr myIP;
  char domainname[32];
/* these flags are used to signal what is already set */
  uint16_t what_is_set;
  pthread_mutex_t semafori[5];
  int numreaders, numwriters;
  struct _netconfig *next;
} netconfig_t;

/* these defines are used as masks for the what_is_set flags set */
#define DNS1			0x0001
#define DNS2			0x0002
#define DOMAINNAME		0x0004
#define NETADDR			0x0008
#define NETBCAST		0x0010
#define UNDER_TEST		0x0020
#define ALL_DONE		0x0040
/* mandatory parameters have higher values for easy checks during heuristics */
#define MY_MAC			0x0800
#define GW_MAC			0x1000
#define NETMASK			0x2000
#define GW_IP			0x4000
#define MY_IP			0x8000

/* these defines are used in the three-entry net_conf array, to identify the
 * current working-on configuration, the last one tried, and, if enabled, a
 * pointer to the profiles list
 */
#define NETCONF_CURR		0
#define NETCONF_BACK		1
#define NETCONF_PROFILES	2

/* function netconf_modify()
 * this function is used to insert or modify a field in a netconfig_t structure,
 * automatically checking if the field is already set, and if the value to
 * be inserted was not already discarded because not working. The mode argument
 * has to be chosen whithin values for the what_is_set flag, while the void*
 * item is just the 
 */
void netconf_modify (int mode, void *item, netconfig_t * net_conf);

/* function netconf_dropcurrent()
 * this function shifts the current configuration in net_conf array to the 
 * last-one-tried position, and reset the current position. This have to be done
 * whenever a configuration is found to be not working. The whats_wrong int, 
 * which has to be chosen between the defines for what_didnt_work, is used
 * to tell what was wrong in previous configuration.
 */
void netconf_dropcurrent (netconfig_t * net_conf, int whats_wrong);

/* function netconf_cmp()
 * this function compares two given net configuration, returning 0 if they have 
 * the same values, -1 else.
 */
int netconf_cmp (netconfig_t * nc1, netconfig_t * nc2);

/* function netconf_init()
 * this function is used to initialise the two-entry net_conf array used to hold
 * work configurations.
 */
void netconf_init (netconfig_t * toinit);

/* function netconf_insertnew()
 * this function inserts a new net configuration proven to be good in the 
 * profile list, so that selfdhcp remembers of it next time.
 */
void netconf_insertnew (netconfig_t * head, netconfig_t * newEntry);

/* function find_profile()
 * given the head entry of the profiles list and a MAC address, this function
 * searches the tree for a profile which has that address as the gateway's
 * address.
 * find_profile() returns a pointer to the profile searched, or NULL if the 
 * profile could not be found or an error occurred.
 */
netconfig_t *find_profile (netconfig_t * head, unsigned char *gwMAC);

/* function xml_parse_profiles()
 * this function parses a children node of the profiles file and stores the
 * values it finds in the proper structure. A pointer to this newly allocated
 * structure is returned, or NULL if any error is found.
 */
netconfig_t *xml_parse_profiles (xmlDocPtr doc, xmlNodePtr cur);

/* function xml_read_profiles()
 * this function reads profiles from an xml file, passed through profile_fname
 * and stores them in a list beginning at head. The integer returned is the
 * number of profiles read, or -1 if the file does not exist.
 */
int xml_read_profiles (char *profile_fname, netconfig_t * head);

/* function write_profilentry()
 * this function writes a single profile entry to the given file.
 */
void write_profilentry (FILE * profile_outfile, netconfig_t * entry);

/* function write_profiles()
 * write_profiles() writes the profiles contained in the list beginning at head
 * to the file passed through profile_fname. The return value is the number of
 * profiles written.
 */
int write_profiles (char *profile_fname, netconfig_t * head);

/* function netconf_print()
 * as easy as it looks like - just prints out to stdout the configuration passed
 * through net_conf.
 */
void netconf_print (netconfig_t * net_conf);

/* function netconf_apply()
 * this function makes active the net configuration previously found. 0 is
 * returned on success, -1 if errors are encountered.
 */
int netconf_apply (char *iface_name, netconfig_t * net_conf);

/* function netconf_unapply()
 * this function resets to zero any network configuration previously applied
 * to the NIC. It returns 0 on success, -1 on errors.
 */
int netconf_unapply (char *iface_name);

/* function selfconf_read()
 * selfconf_read() is used to read an xml configuration file, which contains
 * commonly used parameters. The first argument is the configuration file name,
 * all the others are the parameters to be set (program_flags is not passed
 * because is a global variable). The function returns 1 if succeeds, -1 if the
 * file does not exist.
 */
int selfconf_read (char *conf_fname, char *iface, char *mac,
		   char **prof_fname, int *timeout);

/* function selfconf_write()
 * selfconf_write() writes the current configuration to the file passed through
 * conf_name.
 */
void selfconf_write (char *conf_fname, char *iface, char *mac,
		     char *prof_fname, int *timeout);

/* function change_mac() 
 * This function is used to change the MAC address of the local NIC, in case we
 * don't want to use the real one.
 */
void change_mac (char *mac_addr_str, char *iface_name);

/* wrapped fgets() */
char *Fgets (char *s, int size, FILE * stream);

#endif
