/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_struct.h
 * purposes:	definitions for the main data structures and their functions
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#ifndef __SELFDHCP_STRUCT_H
#define __SELFDHCP_STRUCT_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <libnet.h>
#include <sys/socket.h>
//#include <sys/ethernet.h>
#include <net/if.h>
//#include <netinet/if_ether.h>
#include <netinet/in.h>

/* don't know why people at Sun didn't define this... */
#ifdef __SOLARIS
#define ETH_ALEN ETHER_ADDR_LEN
#endif

#include "selfdhcp_errlib.h"
#include "selfdhcp_pthread.h"

/* extern variables passed from the main function */
extern int program_flags;
extern char *prog_name;

/* these #define are used both to identify the data type in lists, and to access
 * the proper element in the list array */
#define IP_TYPE		1
#define MAC_TYPE	0

/* function mac2str()
 * given a MAC address stored as an array of bytes, it returns the string in
 * the format "xx:xx:xx:xx:xx:xx", or NULL if something was wrong. */
char *mac2str (unsigned char MACaddr_num[ETH_ALEN], char *MACaddr_string);

/* function str2mac()
 * given a strig containing a MAC addr in the "xx:xx:xx:xx:xx:xx" format,
 * returns an unsigned char array containing the numeric form of the address,
 * or NULL if something goes wrong. */
unsigned char *str2mac (char *MACaddr_string,
			unsigned char MACaddr_num[ETH_ALEN]);

/* a sort of meta-structure containing common fields from the IP and MAC lists,
 * plus a void * to hold the real data. It makes code a little more confusing
 * because of long chains of casts, but is a lot more flexible. */
typedef struct lista
{
  void *data;
  struct sublista *relatives;
  int datatype;			/* type of *data, as defined before */
  unsigned int relnum;		/* number of "relatives" */
  unsigned long howmany;	/* how many times we saw this address */
  pthread_mutex_t semafori[5];
  int numreaders, numwriters;
  struct lista *next;		/* numerical order (for IP and MAC) */
  struct lista *nextnum;	/* 'howmany' order */
  struct lista *nextrel;	/* number of relations order */
} lista_t;

/* sublist holds the references to the "other" type of address (IP for MAC and
 * vice versa). */
typedef struct sublista
{
  lista_t *item;
  unsigned long howmany;	/* how many times we saw this couple */
  pthread_mutex_t semafori[5];
  int numreaders, numwriters;
  struct sublista *nextnum;
} sublist_t;

/* function createlists()
 * this function should be used to at first create the struct to hold data. 
 */
lista_t **createlists (void);

/* function coupleinsert()
 * this function is the higher level function for inserting data in the tables;
 * provided two "raw" piece of data (an IP and a MAC), it provides automatically
 * to insert them in the proper structures, passed through the **liste argument.
 * coupleinsert() returns -1 in case of error, or 0 if all was right. */
int coupleinsert (lista_t ** liste, unsigned char newMAC[ETH_ALEN],
		  struct in_addr newIP);

/* function findlevel()
 * findlevel returns the first element of nth 'level' (being level the position)
 * in the count list, with elements with the same counter on the same level).
 * the mode parametere, which should be chosen between the defines below,
 * determines which ordering is used.
 * NULL is returned if the end of the list is reached without finding elements
 * of the required level. */
lista_t *findlevel (lista_t * head, int level, int mode);

/* modes for the findlevel() and findnext() functions */
#define HOW_MANY_TIMES		0	/* howmany order */
#define HOW_MANY_RELATIVES	1	/* relnum order */

/* If this define is used as level, the last level is searched */
#define LAST_ONE		0

/* function findnext()
 * Given an element in a list, findnext() returns the next element with on the
 * same level, or NULL if there are no more elements on that level. Parameter
 * mode works like in findlevel. */
lista_t *findnext (lista_t * element, int mode);

/* function ipcmp()
 * given two in_addr structures ip1 and ip2, it returns -1, 0 or 1 if ip1 is
 * found respectively lesser than, equal or greater than ip2. */
int ipcmp (struct in_addr *ip1, struct in_addr *ip2);

/* function maccmp()
 * given two mac addresses mac1 and mac2, it returns an int lesser than, equal
 * to or greater than zero if mac1 is found respectively lesser than, equal to
 * or greater than mac2. */
int maccmp (unsigned char *mac1, unsigned char *mac2);

/* function insert()
 * this function inserts a generic element in the given list. List type is
 * detected through the internal flag datatype.
 * insert() accepts a pointer to the list head, and a void pointer a previously
 * created chunk of data (the function does not check for the real nature of
 * data, so be careful). A pointer to the newly created (or updated) element is
 * returned in case of success, NULL if something was wrong. */
lista_t *insert (lista_t * head, void *newdata);

/* function search()
 * this function searches in the list pointed from *head for the data passed
 * through data. There are actually two modes of search (see the #define below).
 * search() returns a pointer to the element containing the given data, or NULL 
 * if could not find it. */
lista_t *search (lista_t * head, void *searchdata, int mode);

/* these are the modes for the search() function */
#define FINDITEM_MODE	0	/* item is searched exactly */
#define INSPOINT_MODE	1	/* insertion point is searched instead
				 * (search() stops after finding a higher value)
				 */

/* function sublistinsert()
 * this function is used to insert an item (MAC for IP, and vice-versa), passed
 * through the reldata argumnet, in the "relatives" sublist of the item element. 
 * A pointer to the created or updated record is returned, NULL in case of error.
 */
sublist_t *sublistinsert (lista_t * element, lista_t * reldata);

/* function sublistsearch()
 * searches for a given item reldata in the sublist of the item element.
 * It is returned a pointer to the item found, or NULL if the item was not
 * there. */
sublist_t *sublistsearch (lista_t * element, lista_t * reldata);

/* function relsort()
 * handle the relnum ordering in the list passed through head, after the
 * insertion of the new item element.
 */
void relsort (lista_t * head, lista_t * element);

/* function display_lists()
 * this function is used to print out the MAC and IP lists for debug purposes.
 */
void display_lists (lista_t ** liste);

/* ip2test_t struct is used to hold the list of IP that selfDHCP is testing as 
 * tha gateway's.
 */
typedef struct ip2test
{
  struct in_addr ip;
  pthread_mutex_t semafori[5];
  int numreaders, numwriters;
  struct ip2test *next;
} ip2test_t;

/* function ip2test_insert()
 * inserts a new IP address in the list pointed by head.
 */
void ip2test_insert (ip2test_t * head, struct in_addr ip2ins);

/* function ip2test_search()
 * searches a given IP in the list pointed by head; returns a pointer to the 
 * found IP or NULL if it was not found.
 */
ip2test_t *ip2test_search (ip2test_t * head, struct in_addr ip2search);

/* function ip2test_resetlist()
 * resets a list pointed by head, clearing every IP in the list and freeing the
 * memory allocated.
 */
void ip2test_resetlist (ip2test_t * head);

#endif
