/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_capture.c
 * purposes:	functions for the main capture thread
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "selfdhcp_capture.h"

int in_pos, out_pos;

/* function capture_start()
 * this function is used to start the packet capture sequence. Both the name of
 * the interface to be used and the pointer to the list where MAC and IP
 * addresses are to be stored is bassed through the void** array.
 * libpcap is used both for opening the NIC and for sniffing packets; please 
 * refer to the pcap documentetion for further info (www.tcpdump.org).
 * capture_start() returns -1 if errors are encountered, 0 else.
 */
void
capture_start (void **void_args)
{
  char pcap_errbuf[PCAP_ERRBUF_SIZE] = "";
  pcap_t *capture_session;
//      struct pcap_pkthdr header;
  struct bpf_program my_bpf_filter;
  capture_args_t *args;
//      couple_t *circ_buff;
//      u_char packet;

  args = (capture_args_t *) void_args;
/* The device is opened in promiscuous mode */
  if (NULL == (capture_session = pcap_open_live (args->device, MAX_BYTE,
						 1, 1000, pcap_errbuf)))
    err_quit ("pcap_open_live() failed to open device %s: %s",
	      args->device, pcap_errbuf);
  err_msg (DBG_PROG, "pcap opened device %s", args->device);
/* we compile the filter for the pcap session (see selfdhcp_capture.h) */
  if (-1 ==
      pcap_compile (capture_session, &my_bpf_filter, BPF_FILTER_STR, 1, 0))
    err_quit ("pcap_compile() failed to compile filter: %s",
	      pcap_geterr (capture_session));
/* we apply the compiled filter */
  if (-1 == pcap_setfilter (capture_session, &my_bpf_filter))
    err_quit ("pcap_setfilter() failed to apply filter: %s",
	      pcap_geterr (capture_session));
/* main capture loop */
  while (1)
    {
/* we start the capture loop, with packet_process() as callback function */
      if (-1 ==
	  pcap_loop (capture_session, -1, packet_process, (u_char *) args))
	err_quit ("pcap_loop() failed to start packet capture: %s",
		  pcap_geterr (capture_session));
    }
  pcap_close (capture_session);
  err_msg (DBG_PROG, "exiting capture_start()");
  return;
}

/* function packet_process()
 * this is a standard callback function for pcap_loop(), as specified in the 
 * documentation. This function should be not used directly.
 */
void
packet_process (u_char * args, const struct pcap_pkthdr *header,
		const u_char * packet)
{
  struct libnet_ethernet_hdr *ethernet;
  struct ip *ip;
  uint16_t sport, dport;
  struct arp_hdr *arp;
//      struct libnet_arp_hdr *arp;
  netconfig_t *net_conf;
  capture_args_t *arguments;
  couple_t src_couple, dst_couple;
  ip2test_t *iplist;
  char debug_buffer[20], buf2[20];
  unsigned char bcast_mac[ETH_ALEN] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
  struct in_addr bcast_ip;
  uint32_t ip_buf;

  inet_aton ("255.255.255.255", &bcast_ip);
  arguments = (struct capture_args *) args;
  iplist = arguments->iplist;
  net_conf = arguments->net_conf;
/* The first few bytes of the packet are part of the ethernet header */
  ethernet = (struct libnet_ethernet_hdr *) (packet);
/* Different types of packet need different treatment */
  switch (ntohs (ethernet->ether_type))
    {
/* If it is a plain IP packet, it will contain for sure two valid couples
 * (one for sender and one for receiver) */
    case ETHERTYPE_IP:
      ip = (struct ip *) (packet + sizeof (struct libnet_ethernet_hdr));
      (src_couple.ip_addr).s_addr = (ip->ip_src).s_addr;
      (dst_couple.ip_addr).s_addr = (ip->ip_dst).s_addr;
      memcpy (src_couple.mac_addr, &(ethernet->ether_shost), ETH_ALEN);
      memcpy (dst_couple.mac_addr, &(ethernet->ether_dhost), ETH_ALEN);
/* a test packet is identified if has our MAC and IP, and the gateway's MAC */
      if (net_conf->what_is_set & UNDER_TEST)
	{
	  if (0 == ipcmp (&(dst_couple.ip_addr), &(net_conf->myIP)) &&
	      0 == maccmp (dst_couple.mac_addr, net_conf->myMAC) &&
	      0 == maccmp (src_couple.mac_addr, net_conf->gatewayMAC))
	    {
	      netconf_modify (UNDER_TEST, NULL, net_conf);
	      netconf_modify (ALL_DONE, NULL, net_conf);
	      err_msg (DBG_INFO, "test succeeded!");
	      return;
	    }
	}
/* we have to handle the case of some "interesting" packets, like DNS or DHCP,
 * which could be in TCP or UDP */
      if (ip->ip_p == IPPROTO_TCP || ip->ip_p == IPPROTO_UDP)
	{
/* position of the port is the same in TCP and UDP */
	  sport = ntohs (*((uint16_t *) ((u_char *) ip +
					 sizeof (struct ip))));
	  dport = ntohs (*((uint16_t *) ((u_char *) ip +
					 sizeof (struct ip) +
					 sizeof (uint16_t))));
/* case of a DNS reply */
	  if (sport == DNSPORT)
	    {
	      handle_dns (&src_couple.ip_addr, net_conf);
	    }
/* case of a DNS request */
	  else if (dport == DNSPORT)
	    {
	      handle_dns (&dst_couple.ip_addr, net_conf);
	    }
/* case of SMB broadcast packet */
	  else if (sport == SMBPORT && dport == SMBPORT &&
		   0 == maccmp (bcast_mac, ethernet->ether_dhost) &&
		   program_flags & USE_SMB_BCAST)
	    {
	      handle_smb ((ip->ip_dst).s_addr, net_conf);
	    }
/* case of a BOOTP or DHCP reply (_very_ interesting indeed) */
	  else if (sport == DHCPPORT_SRV && dport == DHCPPORT_CLI)
	    {
	      handle_dhcp (ip, net_conf);
	    }
	}
      break;
/* If we have an ARP packet, what we'll have depends from the type of ARP
   message */
    case ETHERTYPE_ARP:
//                      arp=(struct libnet_arp_hdr *)(packet + sizeof(struct libnet_ethernet_hdr));
      arp = (struct arp_hdr *) (packet + sizeof (struct libnet_ethernet_hdr));
      switch (ntohs (arp->ar_op))
	{
/* if we have a RARP or ARP request, we only have the couple for the sender */
	case ARPOP_REQUEST:
	  (src_couple.ip_addr).s_addr = (*((uint32_t *) arp->ar_spa));
	  memcpy (src_couple.mac_addr, &(ethernet->ether_shost), ETH_ALEN);
	  break;
	case ARPOP_REVREQUEST:
	  (src_couple.ip_addr).s_addr = (*((uint32_t *) arp->ar_spa));
	  memcpy (src_couple.mac_addr, &(ethernet->ether_shost), ETH_ALEN);
	  break;
/* else, if we have a reply, we have both sender and receiver */
	case ARPOP_REPLY:
/* if we got an arp reply with the gateway's MAC as sender, the related IP
 * is for sure the gateway's */
	  if (			/*NULL!=ip2test_search(iplist,
				 *((struct in_addr *)(arp->ar_spa))),*/
	       0 == maccmp (net_conf->gatewayMAC, ethernet->ether_shost))
	    {
//                                              ip2test_resetlist(iplist);
	      ip_buf = (*((uint32_t *) arp->ar_spa));
	      netconf_modify (GW_IP, &ip_buf, net_conf);
	      err_msg (DBG_INFO, "Found gateway IP from ARP reply");
	      memset (src_couple.mac_addr, 0xff, ETH_ALEN);
	      break;
	    }
	  (src_couple.ip_addr).s_addr = (*((uint32_t *) arp->ar_spa));
	  (dst_couple.ip_addr).s_addr = (*((uint32_t *) arp->ar_tpa));
	  memcpy (src_couple.mac_addr, &(ethernet->ether_shost), ETH_ALEN);
	  memcpy (dst_couple.mac_addr, &(ethernet->ether_dhost), ETH_ALEN);
	  break;
	case ARPOP_REVREPLY:
	  (src_couple.ip_addr).s_addr = (*((uint32_t *) arp->ar_spa));
	  (dst_couple.ip_addr).s_addr = (*((uint32_t *) arp->ar_tpa));
	  memcpy (src_couple.mac_addr, &(ethernet->ether_shost), ETH_ALEN);
	  memcpy (dst_couple.mac_addr, &(ethernet->ether_dhost), ETH_ALEN);
	  break;
/* any other ARP packet is junk for us... */
	default:
	  return;
	}
      break;
/* ...just like any other ethernet packet (not IP nor ARP) */
    default:
      return;
    }
/* if we get a packet with our already applied IP, but a MAC which is not ours,
 * we have to immediatly drop the config we're using to avoid IP collision */
  if (net_conf->what_is_set & ALL_DONE)
    if ((0 == ipcmp (&(src_couple.ip_addr), &(net_conf->myIP)) &&
	 0 != maccmp (src_couple.mac_addr, net_conf->myMAC)) ||
	(0 == ipcmp (&(dst_couple.ip_addr), &(net_conf->myIP)) &&
	 0 != maccmp (dst_couple.mac_addr, net_conf->myMAC)))
      {
	err_msg (MSG_WARN, "resetting the network configuration to avoid IP"
		 " collision");
	netconf_unapply (arguments->device);
	netconf_dropcurrent (net_conf, MY_IP);
      }
/* now that we have the yearned couples, it's time to put'em in the circular
 * buffer for the catalog sub-thread, discarding IP or MAC broadcasts */
  if (0 != maccmp (src_couple.mac_addr, bcast_mac) &&
      0 != ipcmp (&(src_couple.ip_addr), &bcast_ip))
    {
      err_msg (DBG_INFO, "captured couple %s %s",
	       mac2str (src_couple.mac_addr, debug_buffer),
	       inet_ntoa (src_couple.ip_addr));
      if (-1 == coupleinsert (arguments->liste, src_couple.mac_addr,
			      src_couple.ip_addr))
	err_msg (MSG_WARN, "coupleinsert() failed", prog_name);
    }
  if (0 != maccmp (dst_couple.mac_addr, bcast_mac) &&
      0 != ipcmp (&(dst_couple.ip_addr), &bcast_ip))
    {
      err_msg (DBG_INFO, "captured couple %s %s",
	       mac2str (dst_couple.mac_addr, debug_buffer),
	       inet_ntoa (dst_couple.ip_addr));
      if (-1 == coupleinsert (arguments->liste, dst_couple.mac_addr,
			      dst_couple.ip_addr))
	err_msg (MSG_WARN, "coupleinsert() failed", prog_name);
    }
  return;
}

/* function handle_dns()
 * handle_dns() is used to set the dns config (which in theory should be a
 * business of the heuristic thread) within the capture thread.
 */
void
handle_dns (struct in_addr *ipaddr, netconfig_t * net_conf)
{

  if (!((net_conf->what_is_set) & DNS1))
    {
      netconf_modify (DNS1, ipaddr, net_conf);
      err_msg (MSG_INFO, "dns1 set to %s", inet_ntoa (net_conf->dns1));
    }
  else if (!((net_conf->what_is_set) & DNS2)
	   && 0 != ipcmp (&(net_conf->dns1), ipaddr))
    {
      netconf_modify (DNS2, ipaddr, net_conf);
      err_msg (MSG_INFO, "dns2 set to %s", inet_ntoa (net_conf->dns2));
    }
  else;
  return;
}

/* function handle_dhcp()
 * given an IP packet containing DHCP or BOOTP informations, handle_dhcp()
 * extracts useful hints such as the netmask, the net address, the gateway IP,
 * etc. Informations are stored directly in net_conf.
 */
void
handle_dhcp (struct ip *ip_pack, netconfig_t * net_conf)
{
  bootp_pack_t *bootp_pack;
  uint8_t *cursor;
  int nexttag, comp;
  struct in_addr ipaddr;

/* we jump up to the starting point of the DHCP or BOOTP payload */
  switch (ip_pack->ip_p)
    {
    case IPPROTO_TCP:
      bootp_pack = (bootp_pack_t *) (ip_pack + sizeof (struct ip)
				     + sizeof (struct tcphdr));
      break;
    case IPPROTO_UDP:
      bootp_pack = (bootp_pack_t *) (ip_pack + sizeof (struct ip)
				     + sizeof (struct udphdr));
      break;
    default:
      break;
    }
  if ((bootp_pack->op) != BOOTREPLY)
    {
      err_msg (DBG_PROG, "got BOOTP or DHCP request packet, discarded");
      return;
    }
  err_msg (DBG_PROG, "got BOOTP or DHCP reply packet");
  cursor = bootp_pack->options;
/* for DHCP options tags refer to rfc1497 and rfc1533 */
/* 255 is the tag for the End Field */
  while (*cursor != 255)
    {
      err_msg (DBG_PROG, "DHCP option tag %d", *(cursor));
      switch (*cursor)
	{
/* Pad Field */
	case 0:
	  nexttag = 1;
	  break;
/* Subnet Mask Field */
	case 1:
	  nexttag = 1 + 1 + 4;
	  ipaddr.s_addr = (uint32_t) * (cursor + 1 + 1);
	  if (!(net_conf->what_is_set) & NETMASK)
	    {
	      netconf_modify (NETMASK, &ipaddr, net_conf);
	      err_msg (MSG_INFO, "netmask from DHCP packet: %s",
		       inet_ntoa (net_conf->netmask));
	    }
	  else
	    {
	      comp = ipcmp (&(net_conf->netmask), &ipaddr);
	      if (comp == 0)
		err_msg (DBG_INFO, "netmask confermed to be %s from DHCP"
			 " packet", inet_ntoa (net_conf->netmask));
/* if the netmask previously found is smaller than the one in DHCP packet, it's
 * better to enlarge it to the new value */
	      else if (comp < 0)
		{
		  err_msg (MSG_INFO, "netmask found %s is smaller than %s"
			   " from DHCP packet, enlarging",
			   inet_ntoa (net_conf->netmask), inet_ntoa (ipaddr));
		  netconf_modify (NETMASK, &ipaddr, net_conf);
		}
	      else
		err_msg (DBG_INFO, "netmask found %s is larger than %s"
			 " from DHCP packet (all right so)",
			 inet_ntoa (net_conf->netmask), inet_ntoa (ipaddr));
	    }
	  break;
/* Router Option (the one we call the gateway) */
	case 3:
	  nexttag = 1 + 1 + 4;
	  ipaddr.s_addr = (uint32_t) * (cursor + 1 + 1);
	  if (!net_conf->what_is_set & GW_IP)
	    {
	      netconf_modify (GW_IP, &ipaddr, net_conf);
	      err_msg (MSG_INFO, "gateway IP from DHCP packet: %s",
		       inet_ntoa (net_conf->gatewayIP));
	    }
	  else
	    {
	      comp = ipcmp (&(net_conf->gatewayIP), &ipaddr);
	      if (comp == 0)
		err_msg (DBG_INFO, "gateway IP confermed to be %s from"
			 " DHCP packet", inet_ntoa (net_conf->gatewayIP));
	      else
		{
		  err_msg (MSG_INFO, "gateway IP found %s is different from"
			   " %s got in DHCP packet, correcting",
			   inet_ntoa (net_conf->gatewayIP),
			   inet_ntoa (ipaddr));
		  netconf_modify (GW_IP, &ipaddr, net_conf);
		}
	    }
	  break;
/* Domain Name Server (DNS) Option */
	case 6:
	  comp = (*(cursor + 1)) / 4;
	  nexttag = 1 + 1 + (comp * 4);
	  for (; comp > 0; comp--)
	    handle_dns ((struct in_addr *) (cursor + 1 + 1 + comp * 4),
			net_conf);
	  break;
/* Domain Name Option */
	case 15:
	  comp = *((cursor + 1));
	  nexttag = 1 + 1 + comp;
	  netconf_modify (DOMAINNAME, (char *) (cursor + 1 + 1), net_conf);
	  err_msg (MSG_INFO, "domain name found in DHCP packet: %s",
		   net_conf->domainname);
	  break;
/* Broadcast Address Option (used to find the network address) */
	case 28:
	  nexttag = 1 + 1 + 4;
	  ipaddr.s_addr = (uint32_t) * (cursor + 1 + 1);
	  err_msg (DBG_INFO, "broadcast found in DHCP packet: %s",
		   inet_ntoa (ipaddr));
/* if we already have the netmask, we directly find the network address */
	  if (net_conf->what_is_set & NETMASK)
	    {
	      ipaddr.s_addr &= ((net_conf->netmask).s_addr);
	      netconf_modify (NETADDR, &ipaddr, net_conf);
	      err_msg (MSG_INFO, "network address calculated from DHCP"
		       " packet: %s", inet_ntoa (net_conf->netaddr));
	    }
/* else we store the net broadcast as is, and let the heuristics find the
 * network address */
	  else
	    netconf_modify (NETBCAST, &ipaddr, net_conf);
	  break;
/* any other DHCP or BOOTP Option is unuseful to our purposes */
	default:
	  break;
	}
      cursor += nexttag;
    }
  return;
}

/* function handle_smb()
 * This function handles the insertion of the informations taken from a smb
 * broadcast packet in the net_conf structure. Right for now it simply calls
 * netconf_modify(). */
void
handle_smb (uint32_t net_bcast, netconfig_t * net_conf)
{
  if (!(net_conf->what_is_set & NETBCAST))
    {
      err_msg (DBG_INFO, "network broadcast from smb: %s",
	       inet_ntoa (*((struct in_addr *) &net_bcast)));
      netconf_modify (NETBCAST, &net_bcast, net_conf);
    }
  else
    if (0 != ipcmp (((struct in_addr *) &net_bcast), &(net_conf->netbcast)))
    {
      err_msg (DBG_INFO, "got two different broadcast addresses");
      err_msg (DBG_INFO, "disabling USE_SMB_BCAST and unsetting netbcast");
      program_flags &= (~USE_SMB_BCAST);
      netconf_modify (NETBCAST, NULL, net_conf);
    }
  return;
}
