/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module: 		selfdhcp_errlib.c
 * purposes:	error and message handling functions
 * reference:	Stevens, Unix network programming (2ed), p.922
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "selfdhcp_errlib.h"

static void
err_doit (int errnoflag, int level, const char *incipit,
	  const char *fmt, va_list ap)
{
  int errno_save = errno, n;
  char buf[MAXLINE + 1];

  snprintf (buf, MAXLINE, "(%s) %s - ", prog_name, incipit);
  n = strlen (buf);
  vsnprintf (buf + n, MAXLINE - n, fmt, ap);
  n = strlen (buf);
  if (errnoflag)
    snprintf (buf + n, MAXLINE - n, ": %s", strerror (errno_save));
/* modified to use the one-bit flags */
  if (program_flags & IS_DAEMON_PROC)
    {
      strcat (buf, "\n");
      syslog (level, buf);
    }
  else
    {
      if (NULL == strstr (buf, "\r"))
	strcat (buf, "\n");
      fflush (stdout);
      fprintf (stderr, "%s", buf);
      fflush (stderr);
    }
}

void
err_ret (const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  err_doit (1, LOG_INFO, "error", fmt, ap);
  va_end (ap);
  return;
}

void
err_sys (const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  err_doit (1, LOG_ERR, "error", fmt, ap);
  va_end (ap);
  exit (1);
}

/* function err_msg()
 * this function has been slightly modified from the original in UNP, so that
 * it can use different "levels" of error messages, including verbose messages
 * and debug informations */
void
err_msg (int level, const char *fmt, ...)
{
  va_list ap;
  char buf[15] = "debug (L";
  int n;

  n = strlen (buf);
  sprintf (buf + n, "%d)", level + 1);
  va_start (ap, fmt);
  switch (level)
    {
    case DBG_INFO:
      if (program_flags & DEBUG_MODE
	  && (program_flags & DEBUG_LEVEL) >= level)
	err_doit (0, LOG_INFO, buf, fmt, ap);
      break;
    case DBG_PROG:
      if (program_flags & DEBUG_MODE
	  && (program_flags & DEBUG_LEVEL) >= level)
	err_doit (0, LOG_INFO, buf, fmt, ap);
      break;
    case DBG_LIB:
      if (program_flags & DEBUG_MODE
	  && (program_flags & DEBUG_LEVEL) >= level)
	err_doit (0, LOG_INFO, buf, fmt, ap);
      break;
    case DBG_LOWLEV:
      if (program_flags & DEBUG_MODE
	  && (program_flags & DEBUG_LEVEL) >= level)
	err_doit (0, LOG_INFO, buf, fmt, ap);
      break;
    case MSG_INFO:
      if (program_flags & VERBOSE_MODE)
	err_doit (0, LOG_INFO, "info", fmt, ap);
      break;
    case MSG_WARN:
      err_doit (0, LOG_INFO, "warning", fmt, ap);
      break;
    case MSG_ERR:
      err_doit (0, LOG_INFO, "error", fmt, ap);
      break;
    default:
      break;
    }
  va_end (ap);
  return;
}

void
err_quit (const char *fmt, ...)
{
  va_list ap;

  va_start (ap, fmt);
  err_doit (0, LOG_ERR, "error", fmt, ap);
  va_end (ap);
  exit (1);
}

/* function display_options()
 * to be used in debug mode to show which options are set. Does not take
 * arguments because the flag variable has global scope.
 */
void
display_options (void)
{
  if (program_flags & ACTIVE_MODE)
    err_msg (DBG_INFO, "active mode flag active");
  if (program_flags & ARP_POISON_ENA)
    err_msg (DBG_INFO, "arp flooding flag active");
  if (program_flags & IS_DAEMON_PROC)
    err_msg (DBG_INFO, "daemon flag active");
  if (program_flags & VERBOSE_MODE)
    err_msg (DBG_INFO, "verbose flag active");
  if (program_flags & RESET_NIC)
    err_msg (DBG_INFO, "reset NIC flag active");
  if (program_flags & DEBUG_MODE)
    err_msg (DBG_INFO, "debug flag active, level %d",
	     (program_flags & DEBUG_LEVEL) + 1);
  if (program_flags & PROFILE_ENABLED)
    err_msg (DBG_INFO, "profile-enabled flag active");
  if (program_flags & TEXTONLY_MODE)
    err_msg (DBG_INFO, "console output flag active");
  return;
}
