/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_heuristics.c
 * purposes:	functions for the heuristic thread and capture sub-threads
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "selfdhcp_heuristics.h"

int
heuristics_start (char *device, lista_t ** liste, int timeout,
		  ip2test_t * iplist, netconfig_t * net_conf,
		  test_hosts_t * hosts)
{
  char lnet_errbuf[LIBNET_ERRBUF_SIZE], buffer[20];
  pack_info_t pack;
  struct ether_addr *_myMAC;
  netconfig_t *prof_found;
  int countdown, step;
//      struct in_addr bcastIP;
  arpp_args_t args;

/* some libnet initialization */
#ifndef NEW_LIBNET_INTERFACE
  err_msg (DBG_PROG, "initializing libnet packet memory");
  if (-1 == libnet_init_packet (LIBNET_MAX_PACKET, &(pack.packetbuff)))
    err_quit ("libnet_init_packet() failed");
  if (NULL ==
      (pack.net_link = libnet_open_link_interface (device, lnet_errbuf)))
    err_quit ("libnet_open_link_interface() failed: %s", lnet_errbuf);
  if (0 == (_myMAC = libnet_get_hwaddr (pack.net_link, device, lnet_errbuf)))
    err_quit ("libnet_get_hwaddr() failed: %s", lnet_errbuf);
#else
  err_msg (DBG_PROG, "initializing libnet stuff");
  if (NULL ==
      (pack.lnet_struct = libnet_init (LIBNET_LINK, device, lnet_errbuf)))
    err_quit ("libnet_init() failed: %s", lnet_errbuf);
  if (0 == (_myMAC = libnet_get_hwaddr (pack.lnet_struct)))
    err_quit ("libnet_get_hwaddr() failed: %s",
	      libnet_geterror (pack.lnet_struct));
#endif
/* these info are the same for every packet we'll inject */
  pack.device = device;
/* build_and_inject() uses a different structure for passing arguments */
  args.device = device;
#ifndef NEW_LIBNET_INTERFACE
  args.packetbuff = pack.packetbuff;
  args.net_link = pack.net_link;
#endif
  net_conf->what_is_set &= 0;
  memcpy (&(net_conf->myMAC), _myMAC, ETH_ALEN);
  net_conf->what_is_set |= MY_MAC;
  err_msg (MSG_INFO, "our MAC address is %s",
	   mac2str (net_conf->myMAC, lnet_errbuf));
  if (NULL == (prof_found = calloc (1, sizeof (netconfig_t))))
    err_quit ("calloc() failed: %s", strerror (errno));

/* if in daemon mode, we have to loop "forever", so countdown is never
 * diminished */
  if (program_flags & IS_DAEMON_PROC)
    step = 0;
/* if not in daemon mode, we loop until we find a configuration or timeout
 * expires */
  else
    step = TIME_GRANULARITY;
/* active mode immediately determines if we are on a known configuration */
  if (program_flags & ACTIVE_MODE)
    {
      err_msg (MSG_INFO, "active search within profiles...");
      if (active_search (net_conf, &pack))
	{
	  if (!((net_conf->what_is_set) & MY_MAC))
	    memcpy (&(net_conf->myMAC), _myMAC, ETH_ALEN);
/* test is done immediately so that we don't have to wait */
	  if (test_config (net_conf, &pack, hosts, &args))
	    {
/* if it's good, do required stuff */
	      err_msg (MSG_INFO, "profile was tested and found valid");
	      if (program_flags & TEXTONLY_MODE)
		netconf_print (net_conf);
	      else
		{
		  err_msg (MSG_INFO, "applying found configuration");
		  if ((net_conf->what_is_set) & MY_MAC &&
		      maccmp (net_conf->myMAC, _myMAC))
		    change_mac (mac2str (net_conf->myMAC, buffer), device);
		  if (-1 == netconf_apply (device, net_conf))
		    err_quit ("could not apply configuration");
		}
/* if in daemon mode, we have to stay up to monitor changes */
	      if (!(program_flags & IS_DAEMON_PROC))
		{
		  err_msg (DBG_INFO, "all is set, exiting");
#ifndef NEW_LIBNET_INTERFACE
		  libnet_destroy_packet (&(pack.packetbuff));
#else
		  libnet_destroy (pack.lnet_struct);
#endif
		  return 1;
		}
	    }
	  else
	    err_msg (MSG_INFO,
		     "profile did not pass the test, going on with heuristics");
	}
    }
/* main heuristics loop */
  for (countdown = timeout; countdown > 0; countdown -= step)
    {
      sleep (TIME_GRANULARITY);
      display_lists (liste);
/* first we find the gateway's MAC */
      if ((net_conf->what_is_set) < (GW_MAC * 2))
	{
	  err_msg (DBG_INFO, "searching gateway MAC");
	  find_gatewayMAC (net_conf, liste);
/* the found gateway's MAC is a known one, we should find it within profiles */
	  if (program_flags & PROFILE_ENABLED &&
	      net_conf->what_is_set & GW_MAC &&
	      NULL != (prof_found = find_profile (&net_conf[NETCONF_PROFILES],
						  net_conf->gatewayMAC)))
	    {
	      err_msg (MSG_INFO,
		       "known network recognized from profiles list");
	      if (prof_found->myMAC == NULL)
		memcpy (prof_found->myMAC, net_conf->myMAC, ETH_ALEN);
	      if (program_flags & DEBUG_MODE)
		netconf_print (prof_found);
	      memcpy (net_conf, prof_found, sizeof (netconfig_t));
	      semaphores_init (net_conf, NETCONF_T);
	      net_conf->what_is_set &=
		(GW_IP | GW_MAC | MY_IP | MY_MAC | NETMASK);
	    }
	}
      else
	err_msg (DBG_INFO, "gateway MAC already set to %s",
		 mac2str (net_conf->gatewayMAC, lnet_errbuf));
/* second, the netmask */
      if (((net_conf->what_is_set) < (NETMASK * 2)) &&
	  ((net_conf->what_is_set) >= GW_MAC))
	{
	  err_msg (DBG_INFO, "searching netmask");
	  find_netmask (net_conf, liste);
	}
      else if (((net_conf->what_is_set) >= GW_MAC) &&
	       ((net_conf[NETCONF_BACK].what_is_set) & NETMASK))
	{
	  err_msg (DBG_INFO, "netmask was not right, adjusting");
	  find_netmask (net_conf, liste);
	}
      else
	{
	  err_msg (DBG_INFO, "netmask already set to %s",
		   inet_ntoa (net_conf->netmask));
	  err_msg (DBG_INFO, "net address already set to %s",
		   inet_ntoa (net_conf->netaddr));
	}
/* third, the gateway's IP */
      if (((net_conf->what_is_set) < (GW_IP * 2)) &&
	  ((net_conf->what_is_set) >= NETMASK))
	{
	  err_msg (DBG_INFO, "searching gateway IP");
	  find_gatewayIP (net_conf, liste, &pack, iplist);
	}
      else
	err_msg (DBG_INFO, "gateway IP already set to %s",
		 inet_ntoa (net_conf->gatewayIP));
/* last thing, we search for a free IP to use */
      if ((net_conf->what_is_set) < (MY_IP * 2) &&
	  (net_conf->what_is_set) >= GW_IP)
	{
	  err_msg (DBG_INFO, "searching a free IP");
	  find_myip (net_conf, liste, &pack);
	}
/* if all is right and set, we test the new configuration */
      if ((net_conf->what_is_set) & GW_IP &&
	  (net_conf->what_is_set) & NETMASK &&
	  (net_conf->what_is_set) & MY_IP)
	{
	  err_msg (MSG_INFO, "testing configuration...");
	  if (test_config (net_conf, &pack, hosts, &args))
	    {
/* if it's good, do required stuff */
	      err_msg (MSG_INFO, "selfdhcp found a valid configuration!");
	      if (program_flags & TEXTONLY_MODE)
		netconf_print (net_conf);
	      else
		{
		  err_msg (MSG_INFO, "applying found configuration");
		  if (-1 == netconf_apply (device, net_conf))
		    err_quit ("could not apply configuration");
		}
/* if in daemon mode, we have to stay up to monitor changes */
	      if (!(program_flags & IS_DAEMON_PROC))
		{
		  err_msg (DBG_INFO, "all is set, exiting");
		  break;
		}
	    }
/* if the configuration found was not good, we drop it without a particular
 * reason and try again hoping for the best */
	  else
	    {
	      err_msg (DBG_INFO, "configuration test failed, try again :-P");
	      netconf_dropcurrent (net_conf, 0);
	    }
	}
    }
#ifndef NEW_LIBNET_INTERFACE
  libnet_destroy_packet (&(pack.packetbuff));
#else
  libnet_destroy (pack.lnet_struct);
#endif
  if ((net_conf->what_is_set) < MY_IP)
    err_quit ("could not find a working configuration");
  else
    return 1;
}

/* function find_myip()
 * find_myip() scans through the IP list, searching for a not yet in use IP
 * address. Of course the address must not only be unused, but also in the
 * netmask. find_myip() updates the current net_conf structure, which must
 * already contain valid gateway and netmask information, with the IP
 * address it found, and returns 1; on failure, 0 is returned and the net_conf
 * structure is not modified.
 */
int
find_myip (netconfig_t * net_conf, lista_t ** liste, pack_info_t * pack)
{
  struct in_addr ip2try;
  uint32_t *_gatewayIP, *_netmask;
//      int replies;

  _gatewayIP = (uint32_t *) & (net_conf->gatewayIP);
  _netmask = (uint32_t *) & (net_conf->netmask);
/* we start scanning from the network address plus one */
  ip2try.s_addr = (net_conf->netaddr).s_addr;
  ip2try.s_addr = htonl (ntohl (ip2try.s_addr) + 1);
/* packets injected are spoofed with the gateway addresses */
  pack->src_mac = net_conf->gatewayMAC;
  pack->src_ip = &(net_conf->gatewayIP);
/* loop until we find a good IP or we tryied every address in the netmask */
  while (is_in_netmask (net_conf, &ip2try))
    {
      if (NULL == search (liste[IP_TYPE], &ip2try, FINDITEM_MODE))
	{
	  err_msg (DBG_INFO, "IP %s is could be free, gonna test",
		   inet_ntoa (ip2try));
	  pack->dst_mac = NULL;
	  pack->dst_ip = &ip2try;
	  arp_req (pack);
	  usleep (2 * REPLY_WAIT_TIMEOUT);
	  if (NULL == search (liste[IP_TYPE], &ip2try, FINDITEM_MODE))
	    {
	      err_msg (MSG_INFO, "IP %s should currently free and usable",
		       inet_ntoa (ip2try));
	      netconf_modify (MY_IP, &ip2try, net_conf);
	      return 1;
	    }
	  err_msg (DBG_INFO, "IP %s is already in use", inet_ntoa (ip2try));
	}
      ip2try.s_addr = htonl (ntohl (ip2try.s_addr) + 1);
    }
  err_msg (MSG_INFO, "there are no free IP in the netmask");
  netconf_dropcurrent (net_conf, NETMASK);
  return 0;
}

/* function find_gatewayMAC()
 * find_gateway() scans the list of MAC addresses to find out if there is one
 * that could be the gateway. In that case, it updates tha net_conf structure
 * and returns 1,else returns 0.
 */
int
find_gatewayMAC (netconfig_t * net_conf, lista_t ** liste)
{
  lista_t *prob_gateway;
  char buffer[20];

/* we check for a possible gateway */
  if (NULL ==
      (prob_gateway = findlevel (liste[MAC_TYPE], 1, HOW_MANY_RELATIVES)))
    {
      err_msg (DBG_INFO, "findlevel did not find a suitable MAC");
      return 0;
    }
  read_reserv (prob_gateway, LISTA_T);
/* we have to discard the MAC found if we already know it's not the gateway's */
  if (net_conf[NETCONF_BACK].what_is_set & GW_MAC &&
      0 == maccmp (net_conf->gatewayMAC, prob_gateway->data))
    {
      read_release (prob_gateway, LISTA_T);
      if (NULL ==
	  (prob_gateway = findnext (prob_gateway, HOW_MANY_RELATIVES)))
	{
	  err_msg (DBG_INFO, "findlevel did not find a suitable MAC");
	  return 0;
	}
      read_reserv (prob_gateway, LISTA_T);
    }
/* it would be a non-sense to have a gateway seen with only one IP */
  if (prob_gateway->relnum < 2)
    {
      err_msg (DBG_INFO, "gateway could not be found");
      read_release (prob_gateway, LISTA_T);
      return 0;
    }
  err_msg (DBG_INFO, "probable gateway found with %d relatives",
	   prob_gateway->relnum);
/* if all is ok, we update the gatewayMAC field and the relative flags */
  netconf_modify (GW_MAC, prob_gateway->data, net_conf);
  err_msg (MSG_INFO, "probable gateway has MAC address %s",
	   mac2str (net_conf->gatewayMAC, buffer));
  read_release (prob_gateway, LISTA_T);
  return 1;
}

/* function find_gatewayIP()
 * function find_gatewayIP() is used, once we have the gateway's MAC and the
 * netmask, to find what the gateway's IP could be.
 * If one can be found, the net_conf entry is updated and 1 is returned; 
 * else 0 is returned.
 */
int
find_gatewayIP (netconfig_t * net_conf, lista_t ** liste, pack_info_t * pack,
		ip2test_t * iplist)
{
  lista_t *cursor, *backcursor;
  sublist_t *sub_cur;
//      struct in_addr *check_ip;
  uint32_t gwIP;
//      int replies;

  ip2test_resetlist (iplist);
/* first, we have to find an IP and a MAC to build spoofed ARP requests */
  read_reserv (liste[MAC_TYPE], LISTA_T);
  cursor = liste[MAC_TYPE]->next;
  read_reserv (cursor, LISTA_T);
  read_release (liste[MAC_TYPE], LISTA_T);
/* of course, they must not be the ones of the gateway itself */
  while (0 == maccmp (cursor->data, net_conf->gatewayMAC))
    {
      backcursor = cursor;
      cursor = cursor->next;
      if (cursor != NULL)
	read_reserv (cursor, LISTA_T);
      read_release (backcursor, LISTA_T);
    }
  if (cursor == NULL)
    {
      err_msg (DBG_INFO, "gateway could not be found");
      return 0;
    }
/* here is the MAC... */
  pack->src_mac = (u_char *) (cursor->data);
/* for the IP, we'll take the first of its relatives */
  read_reserv (cursor->relatives, SUBLIST_T);
  sub_cur = cursor->relatives;
  read_release (cursor, LISTA_T);
  read_reserv (sub_cur->item, LISTA_T);
  cursor = sub_cur->item;
  read_release (sub_cur, SUBLIST_T);
/* ...and here is the IP */
  pack->src_ip = (struct in_addr *) (cursor->data);
  pack->dst_mac = net_conf->gatewayMAC;
  read_release (cursor, LISTA_T);
  err_msg (MSG_INFO, "gonna inject ARP requests with host IP %s",
	   inet_ntoa (*(pack->src_ip)));
/* once we have found both IP and MAC, we can inject the ARP requests.
 * first, we see if the real gateway's IP is listed within its realtives */
  if (NULL ==
      (cursor =
       search (liste[MAC_TYPE], net_conf->gatewayMAC, FINDITEM_MODE)))
    {
      err_msg (MSG_ERR, "couldn't find gateway's MAC in list (?!?)");
      return 0;
    }
  read_reserv (cursor, LISTA_T);
  sub_cur = cursor->relatives;
  read_release (cursor, LISTA_T);
  while (sub_cur != NULL)
    {
      read_reserv (sub_cur, SUBLIST_T);
      cursor = sub_cur->item;
      read_reserv (cursor, LISTA_T);
      err_msg (DBG_INFO, "checking if %s is in netmask",
	       inet_ntoa (*((struct in_addr *) (cursor->data))));
      if (is_in_netmask (net_conf, (struct in_addr *) (cursor->data)))
	{
	  err_msg (MSG_INFO, "testing IP %s as the gateway's",
		   inet_ntoa (*((struct in_addr *) (cursor->data))));
	  pack->dst_ip = (struct in_addr *) (cursor->data);
//                      ip2test_insert(iplist,*((struct in_addr*)(cursor->data)));
	  arp_req (pack);
	}
      read_release (cursor, LISTA_T);
      read_release (sub_cur, SUBLIST_T);
      sub_cur = sub_cur->nextnum;
    }
  usleep (3 * REPLY_WAIT_TIMEOUT);
  if ((net_conf->what_is_set) & GW_IP)
    {
      err_msg (MSG_INFO, "gateway's IP should be %s",
	       inet_ntoa (net_conf->gatewayIP));
      return 1;
    }
/* if not, we try every IP in the netmask */
  gwIP = (net_conf->netaddr).s_addr;
  gwIP = htonl (ntohl (gwIP) + 1);
  err_msg (DBG_INFO, "starting with IP %s for tests",
	   inet_ntoa (*((struct in_addr *) &gwIP)));
/* we loop unltil in netmask or until found the gateway's IP */
  while (is_in_netmask (net_conf, (struct in_addr *) &gwIP) &&
	 !((net_conf->what_is_set) & GW_IP))
    {
      err_msg (MSG_INFO, "testing IP %s as the gateway's    \r",
	       inet_ntoa (*((struct in_addr *) &gwIP)));
      pack->dst_ip = (struct in_addr *) &gwIP;
//              ip2test_insert(iplist,*((struct in_addr *)&gwIP));
      arp_req (pack);
      gwIP = htonl (ntohl (gwIP) + 1);
    }
/* we give some time for the replies to arrive */
  usleep (3 * REPLY_WAIT_TIMEOUT);
  if ((net_conf->what_is_set) & GW_IP)
    {
      err_msg (MSG_INFO, "gateway's IP should be %s",
	       inet_ntoa (net_conf->gatewayIP));
      return 1;
    }
/* is we could not find it, either the gateway's MAC we found is wrong, or,
 * more probably, the netmask is too short */
  err_msg (DBG_INFO, "failed searching the gateway's IP");
  netconf_dropcurrent (net_conf, NETMASK);
  return 0;
}

/* function find_netmask()
 * this function do the heuristic work on the IP list, to find out what the
 * netmask can be. The net_conf struct must already contain a valid MAC entry
 * for the gateway, and is updated with the new netmask. 
 */
void
find_netmask (netconfig_t * net_conf, lista_t ** liste)
{
  struct in_addr netmask, *currIP;
  struct in_addr firstIP, lastIP;
  lista_t *cursor, *backcursor /*, *item_cur */ ;
//      sublist_t *sub_cur, *sub_back;
  int is_first_addr = 1;
//      char buffer[20];

/* we have to find the lower and higher IP in the inner network */
  read_reserv (liste[IP_TYPE], LISTA_T);
  cursor = liste[IP_TYPE]->next;
  read_reserv (cursor, LISTA_T);
  read_release (liste[IP_TYPE], LISTA_T);
  while (cursor != NULL)
    {
/* discarding IP that were found to be "relatives" of the gateway */
      currIP = (struct in_addr *) (cursor->data);
      if (is_not_relative (net_conf->gatewayMAC, currIP, liste))
	{
	  if (is_first_addr)
	    {
	      is_first_addr = 0;
	      err_msg (DBG_PROG, "find_netmask(): %s is first IP",
		       inet_ntoa (*currIP));
	      memcpy (&firstIP, currIP, sizeof (uint32_t));
	      memcpy (&lastIP, currIP, sizeof (uint32_t));
	    }
	  else if (ntohl (currIP->s_addr) < ntohl (firstIP.s_addr))
	    {
	      err_msg (DBG_PROG, "find_netmask(): %s becomes lower IP",
		       inet_ntoa (*currIP));
	      memcpy (&firstIP, currIP, sizeof (uint32_t));
	    }
	  else if (ntohl (currIP->s_addr) > ntohl (lastIP.s_addr))
	    {
	      err_msg (DBG_PROG, "find_netmask(): %s becomes higher IP",
		       inet_ntoa (*currIP));
	      memcpy (&lastIP, currIP, sizeof (uint32_t));
	    }
	  else;
	}
      else
	err_msg (DBG_PROG, "find_netmask(): %s is a gateway relative",
		 inet_ntoa (*currIP));
      backcursor = cursor;
      cursor = cursor->next;
      if (cursor != NULL)
	read_reserv (cursor, LISTA_T);
      read_release (backcursor, LISTA_T);
    }
/* if we have the network brodcast address, we use it as higher inner IP */
//      if(net_conf->what_is_set&NETBCAST)
//              lastIP.s_addr=(net_conf->netbcast).s_addr; 
  if (0 == ipcmp (&firstIP, &lastIP))
    {
      err_msg (MSG_INFO, "not enough inner IPs to calculate netmask");
      return;
    }
  err_msg (DBG_INFO, "lower inner IP is %s", inet_ntoa (firstIP));
  err_msg (DBG_INFO, "higher inner IP is %s", inet_ntoa (lastIP));
/* netmask is calculated as firstIP XNOR currIP */
  netmask.s_addr = ~((firstIP.s_addr) ^ (lastIP.s_addr));
/* some adjust to get rid of spurious 1s */
  netmask_adjust (&netmask);
/* if something was wrong with netmask, we try to enlarge it */
  if (net_conf[NETCONF_BACK].what_is_set & NETMASK &&
      0 == ipcmp (&(net_conf[NETCONF_BACK].netmask), &netmask))
    netmask.s_addr = htonl (ntohl (netmask.s_addr) * 2);
  err_msg (MSG_INFO, "netmask found is %s", inet_ntoa (netmask));
  netconf_modify (NETMASK, &netmask, net_conf);
/* we find the network address starting from a trusted IP (firstIP); lastIP
 * is used as buffer */
  lastIP.s_addr = (netmask.s_addr) & (firstIP.s_addr);
  netconf_modify (NETADDR, &lastIP, net_conf);
  err_msg (MSG_INFO, "network address found is %s",
	   inet_ntoa (net_conf->netaddr));
  return;
}

/* function is_not_relative()
 * is_not_relative() tells wether a given IP called ip2check is within relatives
 * of the gateway or not. If ip2check wasn't found within the gateway's
 * relatives, 1 is returned, if it was or in case of error, is_not_relative()
 * returns 0.
 */
int
is_not_relative (u_char * gwMAC, struct in_addr *ip2check, lista_t ** liste)
{
//      struct in_addr *currIP;
  lista_t *gwEntry, *searchIP;
//      sublist_t *cursor, *backcursor;
/* we get the entry for the gateway's MAC and the IP to search... */
  gwEntry = search (liste[MAC_TYPE], gwMAC, FINDITEM_MODE);
  searchIP = search (liste[IP_TYPE], ip2check, FINDITEM_MODE);
/* ...and sublistsearch() will tell us if they're relatives or not */
  if (NULL == sublistsearch (gwEntry, searchIP))
    return 1;
  return 0;
}

/* function netmask_adjust()
 * this function is used to properly adjust the netmask afer the heuristic
 * calculation (this is because the netmask must be a serie of 1 followed by 0,
 * but they must not mix. netmask_adjust() finds the first 0 in the calculated
 * netmask, and then zeroes all the following bits.
 */
void
netmask_adjust (struct in_addr *netmask)
{
  uint32_t workmask = 0x80000000, new_nm;
  int shift_count = 32;
//      struct in_addr fake;

  new_nm = ntohl (netmask->s_addr);
  while (workmask & new_nm && shift_count > 0)
    {
      workmask /= 2;
      shift_count--;
    }
  while (shift_count > 0)
    {
      new_nm &= (~workmask);
      workmask /= 2;
      shift_count--;
    }
  netmask->s_addr = htonl (new_nm);
}

/* function is_in_netmask()
 * this function tells wether a given IP address ip2check is in the same netmask
 * with the net passed through the net_conf (returns 1) or not (returns 0). 
 */
int
is_in_netmask (netconfig_t * net_conf, struct in_addr *ip2check)
{
  uint32_t checkmask /*, netaddr */ ;
  uint32_t *_netmask, *_netaddr;

  err_msg (DBG_PROG, "%s is_in_netmask()?", inet_ntoa (*ip2check));
  _netaddr = (uint32_t *) & (net_conf->netaddr);
  _netmask = (uint32_t *) & (net_conf->netmask);
/* checkmask is a sort of netmask within the network addr and the ip2check */
  checkmask = ~(*_netaddr ^ (ip2check->s_addr));
  if ((*_netmask) == ((*_netmask) & checkmask))
    {
      err_msg (DBG_PROG, "yes, it is_in_netmask()");
      return 1;
    }
  else
    {
      err_msg (DBG_PROG, "no, it is not");
      return 0;
    }
}

/* function order_dns()
 * order_dns() determines which between the dns found should be the first and
 * which the second.
 */
void
order_dns (netconfig_t * net_conf, lista_t ** liste)
{
  lista_t *dns1, *dns2;
  int dns1_n, dns2_n;

  dns1 = search (liste[IP_TYPE], &(net_conf->dns1.s_addr), FINDITEM_MODE);
  dns2 = search (liste[IP_TYPE], &(net_conf->dns2.s_addr), FINDITEM_MODE);
  dns1_n = dns1->howmany;
  dns2_n = dns2->howmany;
/* if dns2 has been seen more times than dns1, we should switch the two */
  if (dns2_n > dns1_n)
    {
      netconf_modify (DNS1, &(dns2->data), net_conf);
      netconf_modify (DNS2, &(dns1->data), net_conf);
      err_msg (DBG_INFO, "switching dns addresses");
    }
/* else we leave it as is */
  else;
  return;
}

/* function arp_req()
 * arp_req() is used to create and inject an ARP reply. Informations about the
 * addresses and everything are passed through the pack_info structure. If some
 * addresses are found to be NULL, they're filled with 0.
 * The replies are captured and filtered, and net_conf is automatically updated.
 * arp_req() return the number of replies received.
 */
int
arp_req (pack_info_t * packet_info)
{
  u_char bcastMAC[ETH_ALEN] = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };
  uint32_t zeroIP, srcip, dstip;
  int count, bytes;
  char buffer[20];
  u_char arp_pad[ARP_PAD];

  memset (arp_pad, 0, ARP_PAD);
  memset (&zeroIP, 0, sizeof (uint32_t));
  if (packet_info->src_ip == NULL)
    packet_info->src_ip = (struct in_addr *) &zeroIP;
  if (packet_info->dst_ip == NULL)
    packet_info->dst_ip = (struct in_addr *) &zeroIP;
  srcip = (packet_info->src_ip)->s_addr;
  dstip = (packet_info->dst_ip)->s_addr;
  err_msg (DBG_PROG, "src MAC %s IP %s",
	   mac2str (packet_info->src_mac, buffer),
	   inet_ntoa (*(packet_info->src_ip)));
  err_msg (DBG_PROG, "dst MAC %s IP %s", mac2str (bcastMAC, buffer),
	   inet_ntoa (*(packet_info->dst_ip)));
#ifndef NEW_LIBNET_INTERFACE
  err_msg (DBG_PROG, "building ether header");
  if (-1 == libnet_build_ethernet (bcastMAC, packet_info->src_mac,
				   ETHERTYPE_ARP, NULL, 0,
				   packet_info->packetbuff))
    err_quit ("libnet_build_ethernet() failed");
  err_msg (DBG_PROG, "building arp header");
  if (-1 == libnet_build_arp (ARPHRD_ETHER, ETHERTYPE_IP, ETH_ALEN,
			      sizeof (struct in_addr), ARPOP_REQUEST,
			      packet_info->src_mac,
			      (u_char *) & srcip,
			      bcastMAC,
			      (u_char *) & dstip,
			      arp_pad, ARP_PAD,
			      (packet_info->packetbuff) + LIBNET_ETH_H))
    err_quit ("libnet_build_arp() failed");
  err_msg (DBG_PROG, "injecting ARP request to host %s (%d times)",
	   inet_ntoa (*(packet_info->dst_ip)), ARP_REQ_2SEND);
/* packet is injected some times, to get sure it reaches the target */
  for (count = 0; count < ARP_REQ_2SEND; count++)
    {
      usleep (TIME_BETWEEN_SEND);
      if (-1 == (bytes = libnet_write_link_layer (packet_info->net_link,
						  packet_info->device,
						  packet_info->packetbuff,
						  LIBNET_ETH_H +
						  LIBNET_ARP_H + ARP_PAD)))
	err_quit ("libnet_write_link_layer() failed");
    }
  err_msg (DBG_PROG,
	   "libnet_write_link_layer() wrote %d bytes for each packet", bytes);
#else
  err_msg (DBG_PROG, "building arp header");
  if (-1 == libnet_build_arp (ARPHRD_ETHER, ETHERTYPE_IP, ETH_ALEN,
			      sizeof (struct in_addr), ARPOP_REQUEST,
			      packet_info->src_mac,
			      (u_char *) & srcip,
			      bcastMAC,
			      (u_char *) & dstip,
			      arp_pad, ARP_PAD, packet_info->lnet_struct, 0))
    err_quit ("libnet_build_arp() failed");
  err_msg (DBG_PROG, "building ether header");
  if (-1 == libnet_build_ethernet (bcastMAC, packet_info->src_mac,
				   ETHERTYPE_ARP, NULL, 0,
				   packet_info->lnet_struct, 0))
    err_quit ("libnet_build_ethernet() failed");
  err_msg (DBG_PROG, "injecting ARP request to host %s (%d times)",
	   inet_ntoa (*(packet_info->dst_ip)), ARP_REQ_2SEND);
/* packet is injected some times, to get sure it reaches the target */
  for (count = 0; count < ARP_REQ_2SEND; count++)
    {
      usleep (TIME_BETWEEN_SEND);
      if (-1 == (bytes = libnet_write (packet_info->lnet_struct)))
	err_quit ("libnet_write() failed: %s",
		  libnet_geterror (packet_info->lnet_struct));
    }
  err_msg (DBG_PROG, "libnet_write() wrote %d bytes for each packet", bytes);

#endif
  usleep (REPLY_WAIT_TIMEOUT);
  return 1;
}

/* function test_config()
 * test_config() is used to create and inject some fake DNS requests.
 * 1 is returned if at least one of these request got reply, 0 if nobody
 * answered.
 */
int
test_config (netconfig_t * net_conf, pack_info_t * packet_info,
	     test_hosts_t * hosts, arpp_args_t * args)
{
  uint32_t srcip, dstip, *testhosts, bcastIP = 0xFFFFFFFF;
  int count, pad_len;
  char buffer[20], payload[] = ICMP_PAD;

  usleep (10 * REPLY_WAIT_TIMEOUT);
  err_msg (DBG_PROG, "gonna inject %d ICMP echo request(s)", hosts->num);
  testhosts = hosts->test_IPs;
  pad_len = strlen (payload);
  srcip = (net_conf->myIP).s_addr;
/* some gratuitous ARP requests are sent to advertise our presence */
  err_msg (DBG_PROG, "injecting some gratuitous ARP requests");
  for (count = 0; count <= ARP_REQ_2SEND; count++)
    build_and_inject (args, srcip, bcastIP, net_conf->myMAC, ARPOP_REPLY);
  netconf_modify (UNDER_TEST, NULL, net_conf);
#ifndef NEW_LIBNET_INTERFACE
/* with libnet 1.0 ethernet header is built just once, because is always the
 * same and the construction order is not important */
  err_msg (DBG_PROG, "building ether header");
  if (-1 == libnet_build_ethernet (net_conf->gatewayMAC, net_conf->myMAC,
				   ETHERTYPE_IP, NULL, 0,
				   packet_info->packetbuff))
    err_quit ("libnet_build_ethernet() failed");
#endif
/* for every host in the test host array,  */
  for (count = 0; count <= hosts->num && !(net_conf->what_is_set & ALL_DONE);
       count++)
    {
      dstip = htonl (testhosts[count]);
      err_msg (DBG_INFO, "%d test host %s", count + 1,
	       inet_ntoa (*((struct in_addr *) &dstip)));
      err_msg (DBG_PROG, "src MAC %s IP %s",
	       mac2str (net_conf->myMAC, buffer), inet_ntoa (net_conf->myIP));
      err_msg (DBG_PROG, "dst MAC %s IP %s",
	       mac2str (net_conf->gatewayMAC, buffer),
	       inet_ntoa (*((struct in_addr *) &dstip)));
#ifndef NEW_LIBNET_INTERFACE
      err_msg (DBG_LIB, "building IP header");
      if (-1 == libnet_build_ip (ICMP_ECHO_H + pad_len, 0, 242, 0x4000, 48,
				 IPPROTO_ICMP, srcip, dstip, NULL, 0,
				 (packet_info->packetbuff) + LIBNET_ETH_H))
	err_msg (MSG_ERR, "libnet_build_ip() failed");
      err_msg (DBG_LIB, "building ICMP header");
      if (-1 ==
	  libnet_build_icmp_echo (ICMP_ECHO, 0, 242, 5, ICMP_PAD, pad_len,
				  (packet_info->packetbuff) + LIBNET_ETH_H +
				  LIBNET_IP_H))
	err_msg (MSG_ERR, "libnet_build_icmp_echo() failed");
      err_msg (DBG_LIB, "calculating checksum");
      if (-1 == libnet_do_checksum ((packet_info->packetbuff) + LIBNET_ETH_H,
				    IPPROTO_IP, LIBNET_IP_H) ||
	  -1 == libnet_do_checksum ((packet_info->packetbuff) + LIBNET_ETH_H,
				    IPPROTO_ICMP,
				    LIBNET_ICMP_ECHO_H + pad_len))
	err_msg (MSG_ERR, "libnet_do_checksum() failed");
      err_msg (DBG_LIB, "injecting test packet");
      if (-1 == libnet_write_link_layer (packet_info->net_link,
					 packet_info->device,
					 packet_info->packetbuff,
					 LIBNET_ETH_H + LIBNET_IP_H +
					 LIBNET_ICMP_ECHO_H + pad_len))
	err_msg (MSG_ERR, "libnet_write_link_layer() failed");
#else
      err_msg (DBG_LIB, "building ICMP header");
      if (-1 ==
	  libnet_build_icmpv4_echo (ICMP_ECHO, 0, 0, 242, 5, ICMP_PAD,
				    pad_len, packet_info->lnet_struct, 0))
	err_msg (MSG_ERR, "libnet_build_icmpv4_echo() failed: %s",
		 libnet_geterror (packet_info->lnet_struct));
      err_msg (DBG_LIB, "building IP header");
      if (-1 ==
	  libnet_build_ipv4 (LIBNET_ICMPV4_ECHO_H + pad_len, 0, 242, 0x4000,
			     48, IPPROTO_ICMP, 0, srcip, dstip, NULL, 0,
			     packet_info->lnet_struct, 0))
	err_msg (MSG_ERR, "libnet_build_ipv4() failed: %s",
		 libnet_geterror (packet_info->lnet_struct));
      err_msg (DBG_PROG, "building ether header");
      if (-1 == libnet_build_ethernet (net_conf->gatewayMAC, net_conf->myMAC,
				       ETHERTYPE_IP, NULL, 0,
				       packet_info->lnet_struct, 0))
	err_quit ("libnet_build_ethernet() failed: %s",
		  libnet_geterror (packet_info->lnet_struct));
      err_msg (DBG_LIB, "injecting test packet");
      if (-1 == libnet_write (packet_info->lnet_struct))
	err_msg (MSG_ERR, "libnet_write_link_layer() failed: %s",
		 libnet_geterror (packet_info->lnet_struct));
#endif
      usleep (REPLY_WAIT_TIMEOUT);
    }
/* we give some more time for replies to come back */
  usleep (10 * REPLY_WAIT_TIMEOUT);
  if (net_conf->what_is_set & ALL_DONE)
    {
      err_msg (DBG_INFO, "test succeeded!");
      return 1;
    }
  netconf_modify (UNDER_TEST, NULL, net_conf);
  err_msg (DBG_INFO, "test failed");
  return 0;
}

/* function active_search()
 * this function actively sends ARP request to the gateway of every profile
 * loaded. If one gives reply, the search is stopped and the profile is copied
 * into the current network configuration. 1 is returned if a profiles fitted,
 * 0 else.
 */
int
active_search (netconfig_t * net_conf, pack_info_t * pack)
{
  netconfig_t *cursor;
  int num = 1;

  cursor = &(net_conf[NETCONF_PROFILES]);
  cursor = cursor->next;
  while (cursor != NULL)
    {
      err_msg (DBG_INFO, "trying profile number %d", num++);
//              netconf_modify(GW_MAC,cursor->gatewayMAC,net_conf);
      memcpy (net_conf->gatewayMAC, cursor->gatewayMAC, ETH_ALEN);
      net_conf->what_is_set |= GW_MAC;
      pack->dst_mac = cursor->gatewayMAC;
      pack->src_mac = net_conf->myMAC;
      pack->src_ip = &(cursor->myIP);
      pack->dst_ip = &(cursor->gatewayIP);
      arp_req (pack);
      usleep (REPLY_WAIT_TIMEOUT);
      if ((net_conf->what_is_set) & GW_IP)
	{
	  err_msg (MSG_INFO,
		   "configuration found from profiles list, going for"
		   " test");
	  memcpy (net_conf, cursor, sizeof (netconfig_t));
	  semaphores_init (net_conf, NETCONF_T);
	  net_conf->what_is_set &=
	    (GW_IP | GW_MAC | MY_IP | MY_MAC | NETMASK);
	  return 1;
	}
      cursor = cursor->next;
    }
  err_msg (MSG_INFO, "no profile was appliable, going on with heuristics");
  return 0;
}
