/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_profiles.h
 * purposes:	functions for the profiles and network configurations management
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

#include "selfdhcp_profiles.h"

/* function netconf_modify()
 * this function is used to insert or modify a field in a netconfig_t structure,
 * automatically checking if the field is already set, and if the value to
 * be inserted was not already discarded because not working. The mode argument
 * has to be chosen whithin values for the what_is_set flag, while the void*
 * item is just the 
 */
void
netconf_modify (int mode, void *item, netconfig_t * net_conf)
{
  char buffer[20];

  err_msg (DBG_LIB, "netconf_modify(): locking net_conf");
  write_reserv (net_conf, NETCONF_T);
/* these two modes do not modify data, but flags only */
  if (mode == ALL_DONE || mode == UNDER_TEST)
    {
      err_msg (DBG_PROG, "netconf_modify(): setting/unsetting mode");
      net_conf->what_is_set ^= mode;
      err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
      write_release (net_conf, NETCONF_T);
      return;
    }
/* if the provided item is already set, and not already tested, we skip */
  if (net_conf->what_is_set & mode)
    {
      err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
      write_release (net_conf, NETCONF_T);
      return;
    }
/* else if it's not set, we have to check if failed in a previous net config */
  if ((net_conf[NETCONF_BACK]).what_is_set & mode)
    {
      switch (mode)
	{
	case GW_MAC:
	  if (0 == maccmp ((unsigned char *) item,
			   (net_conf[NETCONF_BACK]).gatewayMAC))
	    {
	      err_msg (DBG_INFO, "gateway MAC %s discarded because did not"
		       " work previously",
		       mac2str ((unsigned char *) item, buffer));
	      err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
	      write_release (net_conf, NETCONF_T);
	      return;
	    }
	case NETMASK:
	  if (0 == ipcmp ((struct in_addr *) item,
			  &((net_conf[NETCONF_BACK]).gatewayIP)))
	    {
	      err_msg (DBG_INFO, "netmaks %s discarded because did not"
		       " work previously",
		       inet_ntoa (*(struct in_addr *) item));
	      err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
	      write_release (net_conf, NETCONF_T);
	      return;
	    }
	case GW_IP:
	  if (0 == ipcmp ((struct in_addr *) item,
			  &((net_conf[NETCONF_BACK]).gatewayIP)))
	    {
	      err_msg (DBG_INFO, "gateway IP %s discarded because did not"
		       " work previously",
		       inet_ntoa (*(struct in_addr *) item));
	      err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
	      write_release (net_conf, NETCONF_T);
	      return;
	    }
	case MY_IP:
	  if (0 == ipcmp ((struct in_addr *) item,
			  &((net_conf[NETCONF_BACK]).myIP)))
	    {
	      err_msg (DBG_INFO, "IP %s discarded because did not"
		       " work previously",
		       inet_ntoa (*(struct in_addr *) item));
	      err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
	      write_release (net_conf, NETCONF_T);
	      return;
	    }
	default:
	  break;
	}
    }
/* if not already set and not previously discarded, we update the structure */
  switch (mode)
    {
    case GW_MAC:
      memcpy (net_conf->gatewayMAC, item, ETH_ALEN);
      net_conf->what_is_set |= GW_MAC;
      break;
    case NETMASK:
      (net_conf->netmask).s_addr = *((uint32_t *) item);
      net_conf->what_is_set |= NETMASK;
      break;
    case GW_IP:
      (net_conf->gatewayIP).s_addr = *((uint32_t *) item);
      net_conf->what_is_set |= GW_IP;
      break;
    case MY_IP:
      (net_conf->myIP).s_addr = *((uint32_t *) item);
      net_conf->what_is_set |= MY_IP;
      break;
    case NETADDR:
      (net_conf->netaddr).s_addr = *((uint32_t *) item);
      net_conf->what_is_set |= NETADDR;
      break;
    case NETBCAST:
      if (item == NULL)
	net_conf->what_is_set &= (~NETBCAST);
      else
	{
	  (net_conf->netbcast).s_addr = *((uint32_t *) item);
	  net_conf->what_is_set |= NETBCAST;
	}
      break;
    case DNS1:
      (net_conf->dns1).s_addr = *((uint32_t *) item);
      net_conf->what_is_set |= DNS1;
      break;
    case DNS2:
      (net_conf->dns2).s_addr = *((uint32_t *) item);
      net_conf->what_is_set |= DNS2;
      break;
    case DOMAINNAME:
      strncpy (net_conf->domainname, (char *) item, 32);
      net_conf->what_is_set |= DOMAINNAME;
      break;
    default:
      err_quit ("invalid %d argument in netconf_modify() - this is"
		" probably a bug");
    }
  err_msg (DBG_LIB, "netconf_modify(): unlocking net_conf");
  write_release (net_conf, NETCONF_T);
  return;
}

/* function netconf_dropcurrent()
 * this function shifts the current configuration in net_conf array to the 
 * last-one-tried position, and reset the current position. This have to be done
 * whenever a configuration is found to be not working. The whats_wrong int, 
 * which has to be chosen between the defines for what_is_set, is used
 * to tell what was wrong in previous configuration.
 */
void
netconf_dropcurrent (netconfig_t * net_conf, int whats_wrong)
{
  err_msg (DBG_LIB, "netconf_dropcurrent(): locking net_conf[NETCONF_CURR]");
  write_reserv (&(net_conf[NETCONF_CURR]), NETCONF_T);
  err_msg (DBG_LIB, "netconf_dropcurrent(): locking net_conf[NETCONF_BACK]");
  write_reserv (&(net_conf[NETCONF_BACK]), NETCONF_T);
  memcpy (&net_conf[NETCONF_CURR], &net_conf[NETCONF_BACK],
	  sizeof (netconfig_t));
  memset (&net_conf[NETCONF_CURR], '\0', sizeof (netconfig_t));
  if (((net_conf[NETCONF_BACK]).what_is_set) & MY_MAC)
    memcpy ((net_conf[NETCONF_CURR]).myMAC, (net_conf[NETCONF_BACK]).myMAC,
	    ETH_ALEN);
  net_conf[NETCONF_BACK].what_is_set |= whats_wrong;
  err_msg (DBG_LIB,
	   "netconf_dropcurrent(): unlocking net_conf[NETCONF_BACK]");
  write_release (&(net_conf[NETCONF_BACK]), NETCONF_T);
  err_msg (DBG_LIB,
	   "netconf_dropcurrent(): initializing net_conf[NETCONF_CURR]"
	   " seaphores");
  semaphores_init (&(net_conf[NETCONF_CURR]), NETCONF_T);
  return;
}

/* function netconf_cmp()
 * this function compares two given net configuration, returning 0 if they have 
 * the same values, non-zero values else.
 */
int
netconf_cmp (netconfig_t * nc1, netconfig_t * nc2)
{
  return memcmp (nc1, nc2, sizeof (netconfig_t));
}

/* function netconf_init()
 * this function is used to initialise the three-entry net_conf array used to 
 * hold work configurations and profiles.
 */
void
netconf_init (netconfig_t * toinit)
{
  memset (toinit, '\0', 3 * sizeof (netconfig_t));
  semaphores_init (toinit, NETCONF_T);
  return;
}

/* function netconf_insertnew()
 * this function inserts a new net configuration proven to be good in the 
 * profile list, so that selfdhcp remembers of it next time.
 */
void
netconf_insertnew (netconfig_t * head, netconfig_t * newEntry)
{
  netconfig_t *cursor, *backcursor;
  cursor = head->next;
  backcursor = cursor;
  while (cursor != NULL)
    {
      backcursor = cursor;
      cursor = cursor->next;
    }
  if (backcursor == NULL)
    head->next = newEntry;
  else
    {
      newEntry->next = backcursor->next;
      backcursor->next = newEntry;
    }
  return;
}

/* function find_profile()
 * given the head entry of the profiles list and a MAC address, this function
 * searches the tree for a profile which has that address as the gateway's
 * address.
 * find_profile() returns a pointer to the profile searched, or NULL if the 
 * profile could not be found or an error occurred.
 */
netconfig_t *
find_profile (netconfig_t * head, unsigned char *gwMAC)
{
  netconfig_t *cursor;
  char buffer[20];
  int num = 0;

  cursor = head->next;
  while (cursor != NULL)
    {
      err_msg (DBG_PROG, "reading profile %d, gateway MAC %s",
	       ++num, mac2str (cursor->gatewayMAC, buffer));
      if (0 == maccmp (cursor->gatewayMAC, gwMAC))
	break;
      cursor = cursor->next;
    }
  return cursor;
}

/* function xml_parse_profiles()
 * this function parses a children node of the profiles file and stores the
 * values it finds in the proper structure. A pointer to this newly allocated
 * structure is returned, or NULL if any error is found.
 */
netconfig_t *
xml_parse_profiles (xmlDocPtr doc, xmlNodePtr cur)
{
  xmlChar *key;
  netconfig_t netconf_buff, *newEntry;

  memset (&netconf_buff, 0, sizeof (netconfig_t));
  cur = cur->xmlChildrenNode;
/* while we got children elements of the current profile, we check one by one
 * the content */
  while (cur != NULL)
    {
/* gateway MAC */
      if ((!xmlStrcmp (cur->name, (const xmlChar *) "gatewayMAC")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 2);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (NULL == str2mac ((char *) key, netconf_buff.gatewayMAC))
	    {
	      err_msg (MSG_ERR, "error reading gateway's MAC %s",
		       (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= GW_MAC;
	  err_msg (DBG_INFO, "read gateway MAC %s",
		   mac2str (netconf_buff.gatewayMAC, (char *) key));
	  xmlFree (key);
	}
/* gateway IP */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "gatewayIP")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (0 == inet_aton ((char *) key, &(netconf_buff.gatewayIP)))
	    {
	      err_msg (MSG_ERR, "invalid gateway's IP %s", (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= GW_IP;
	  err_msg (DBG_INFO, "read gateway IP %s",
		   inet_ntoa (netconf_buff.gatewayIP));
	  xmlFree (key);
	}
/* netmask */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "netmask")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (0 == inet_aton ((char *) key, &(netconf_buff.netmask)))
	    {
	      err_msg (MSG_ERR, "invalid netmask %s", (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= NETMASK;
	  err_msg (DBG_INFO, "read netmask %s",
		   inet_ntoa (netconf_buff.netmask));
	  xmlFree (key);
	}
/* network address */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "netaddr")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (0 == inet_aton ((char *) key, &(netconf_buff.netaddr)))
	    {
	      err_msg (MSG_ERR, "invalid network address %s", (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= NETADDR;
	  err_msg (DBG_INFO, "read netaddr %s",
		   inet_ntoa (netconf_buff.netaddr));
	  xmlFree (key);
	}
/* domain name */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "domainName")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  strncpy (netconf_buff.domainname, (char *) key, 49);
	  netconf_buff.what_is_set &= NETADDR;
	  err_msg (DBG_INFO, "read domain name \"%s\"",
		   netconf_buff.domainname);
	  xmlFree (key);
	}
/* ourselves MAC */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "myMAC")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 2);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (NULL == str2mac ((char *) key, netconf_buff.myMAC))
	    {
	      err_msg (MSG_ERR, "error reading ourselves MAC %s",
		       (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= MY_MAC;
	  err_msg (DBG_INFO, "read ourselves MAC %s",
		   mac2str (netconf_buff.myMAC, (char *) key));
	  xmlFree (key);
	}
/* ourselves IP */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "myIP")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (0 == inet_aton ((char *) key, &(netconf_buff.myIP)))
	    {
	      err_msg (MSG_ERR, "invalid ourselves IP %s", (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= MY_IP;
	  err_msg (DBG_INFO, "read ourselves IP %s",
		   inet_ntoa (netconf_buff.myIP));
	  xmlFree (key);
	}
/* DNS1 */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "dns1")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (0 == inet_aton ((char *) key, &(netconf_buff.dns1)))
	    {
	      err_msg (MSG_ERR, "invalid DNS1 IP %s", (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= DNS1;
	  err_msg (DBG_INFO, "read DNS1 %s", inet_ntoa (netconf_buff.dns1));
	  xmlFree (key);
	}
/* DNS2 */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "dns2")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (0 == inet_aton ((char *) key, &(netconf_buff.dns2)))
	    {
	      err_msg (MSG_ERR, "invalid DNS2 IP %s", (char *) key);
	      return NULL;
	    }
	  netconf_buff.what_is_set &= DNS2;
	  err_msg (DBG_INFO, "read DNS2 %s", inet_ntoa (netconf_buff.dns2));
	  xmlFree (key);
	}
/* comment (not used) */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "comment")))
	{
	  err_msg (DBG_INFO, "read comment, not used");
	}
/* junk */
      else;
/* then we get to the next element */
      cur = cur->next;
    }
/* if all was right, a new struct is allocated, and the content is copied from
 * the buffer */
  if (NULL == (newEntry = calloc (1, sizeof (netconfig_t))))
    err_sys ("calloc() failed");
  memcpy (newEntry, &netconf_buff, sizeof (netconfig_t));
  newEntry->next = NULL;
  return newEntry;
}

/* function xml_read_profiles()
 * this function reads profiles from an xml file, passed through profile_fname
 * and stores them in a list beginning at head. The integer returned is the
 * number of profiles read, or -1 if the file does not exist.
 */
int
xml_read_profiles (char *profile_fname, netconfig_t * net_conf)
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  netconfig_t *newEntry, *backentry;
  int count = 0;

  backentry = &(net_conf[NETCONF_PROFILES]);
/* xml file opening */
  if (NULL == (doc = xmlParseFile (profile_fname)))
    {
/* if the file does not exist, it will be created and filled with found
 * configurations if also the -u (--update) option is set */
      if (errno == ENOENT)
	return -1;
      else
	err_quit ("could not open profiles file %s: %s", profile_fname,
		  strerror (errno));
    }
/* we check for the file being not empty */
  if (NULL == (cur = xmlDocGetRootElement (doc)))
    {
      err_msg (MSG_WARN, "empty profiles file %s", profile_fname);
      xmlFreeDoc (doc);
      return 0;
    }
/* we check for the file being of proper type */
  if (xmlStrcmp (cur->name, (const xmlChar *) "NetworkCollection"))
    {
      xmlFreeDoc (doc);
      err_quit ("profiles file %s of the wrong type, root node is not"
		" NetworkCollection", profile_fname);
    }
  cur = cur->xmlChildrenNode;
/* loop untile there are children elements */
  while (cur != NULL)
    {
      if ((!xmlStrcmp (cur->name, (const xmlChar *) "NetworkParameters")))
	{
	  err_msg (DBG_INFO, "reading profile n%d", ++count);
	  if (NULL == (newEntry = xml_parse_profiles (doc, cur)))
	    {
	      err_msg (MSG_ERR, "error while reding profile n%d", count);
	      backentry->next = NULL;
	    }
/* if the read was successful, we insert the newly created struct in the list */
	  else
	    {
	      backentry->next = newEntry;
	      backentry = newEntry;
	      err_msg (DBG_INFO, "profile loaded");
	    }
	}
      err_msg (DBG_PROG, "going for next profile...");
      cur = cur->next;
    }
  err_msg (DBG_PROG, "closing profiles file");
  xmlFreeDoc (doc);
  return count;
}

/* function write_profilentry()
 * this function writes a single profile entry to the given file.
 */
void
write_profilentry (FILE * profile_outfile, netconfig_t * entry)
{
  char buffer[32];
  struct in_addr ipaddr;

  inet_aton ("0.0.0.0", &ipaddr);
  fprintf (profile_outfile, "  <NetworkParameters>\n");
  fprintf (profile_outfile, "    <gateway MAC>%s</gatewayMAC>\n",
	   mac2str (entry->gatewayMAC, buffer));
  if (0 != ipcmp (&(entry->gatewayIP), &ipaddr))
    fprintf (profile_outfile, "    <gatewayIP>%s</gatewayIP>\n",
	     inet_ntoa (entry->gatewayIP));
  if (0 != ipcmp (&(entry->netmask), &ipaddr))
    fprintf (profile_outfile, "    <netmask>%s</netmask>\n",
	     inet_ntoa (entry->netmask));
  if (0 != ipcmp (&(entry->netaddr), &ipaddr))
    fprintf (profile_outfile, "    <netaddr>%s</netaddr>\n",
	     inet_ntoa (entry->netaddr));
  if (strlen (entry->domainname))
    fprintf (profile_outfile, "    <domainName>%s</domainName>\n",
	     entry->domainname);
  if (0 != ipcmp (&(entry->myIP), &ipaddr))
    fprintf (profile_outfile, "    <myMAC>%s</myMAC>\n",
	     mac2str (entry->myMAC, buffer));
  if (0 != ipcmp (&(entry->myIP), &ipaddr))
    fprintf (profile_outfile, "    <myIP>%s</myIP>\n",
	     inet_ntoa (entry->myIP));
  if (0 != ipcmp (&(entry->dns1), &ipaddr))
    fprintf (profile_outfile, "    <dns1>%s</dns1>\n",
	     inet_ntoa (entry->dns1));
  if (0 != ipcmp (&(entry->dns2), &ipaddr))
    fprintf (profile_outfile, "    <dns2>%s</dns2>\n",
	     inet_ntoa (entry->dns2));
  fprintf (profile_outfile, "  </NetworkParameters>\n");
  return;
}

/* function write_profiles()
 * write_profiles() writes the profiles contained in the list beginning at head
 * to the file passed through profile_fname. The return value is the number of
 * profiles written.
 */
int
write_profiles (char *profile_fname, netconfig_t * head)
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  FILE *doc_;
  netconfig_t *cursor;
  int count = 0;

  cursor = head;
/* xml file opening */
  if (NULL == (doc = xmlParseFile (profile_fname)))
    {
/* if file does not exist, we create it */
      if (ENOENT == errno);
      err_quit ("could not open profiles file %s: %s", profile_fname,
		strerror (errno));
    }
/* we check for the file being not empty */
  else
    {
      if (NULL == (cur = xmlDocGetRootElement (doc)))
	{
	  err_msg (MSG_WARN, "empty profiles file %s", profile_fname);
	  xmlFreeDoc (doc);
	}
/* we check for the file being of proper type */
      if (xmlStrcmp (cur->name, (const xmlChar *) "NetworkCollection"))
	{
	  xmlFreeDoc (doc);
	  err_quit ("profiles file %s of the wrong type, root node is not"
		    " NetworkCollection", profile_fname);
	}
      xmlFreeDoc (doc);
    }
  if (NULL == (doc_ = fopen (profile_fname, "w")))
    err_quit ("fopen() on file %s failed: %s", profile_fname,
	      strerror (errno));
  fprintf (doc_, "<?xml version=\"\1.0\"?>\n");
  fprintf (doc_, "<NetworkCollection>\n");
/* loop untile there are elements in the list */
  while (cursor != NULL)
    {
      err_msg (DBG_INFO, "writing profile n%d", ++count);
      write_profilentry (doc_, cursor);
      cursor = cursor->next;
    }
  fprintf (doc_, "</NetworkCollection>\n");
  fclose (doc_);
  return count;
}

/* function netconf_print()
 * as easy as it looks like - just prints out to stdout the configuration passed
 * through net_conf.
 */
void
netconf_print (netconfig_t * net_conf)
{
  char buffer[20];

/* printf() is used, so that the output can be redirected */
  if (net_conf->domainname != NULL)
    printf ("Domain name:\t%s\n", net_conf->domainname);
  else
    printf ("Domain name unknown");
  printf ("Gateway MAC:\t%s\n", mac2str (net_conf->gatewayMAC, buffer));
  printf ("Gateway IP:\t%s\n", inet_ntoa (net_conf->gatewayIP));
  printf ("Netmask:\t%s\n", inet_ntoa (net_conf->netmask));
  printf ("My IP:\t\t%s\n", inet_ntoa (net_conf->myIP));
  printf ("My MAC:\t\t%s\n", mac2str (net_conf->myMAC, buffer));
  printf ("DNS1:\t\t%s\n", inet_ntoa (net_conf->dns1));
  printf ("DNS2:\t\t%s\n", inet_ntoa (net_conf->dns2));
  return;
}

/* function netconf_apply()
 * this function makes active the net configuration previously found. 0 is
 * returned on success, -1 if errors are encountered.
 */
int
netconf_apply (char *iface_name, netconfig_t * net_conf)
{
  struct ifreq ifr;
  struct rtentry rtent;
  struct sockaddr_in *p = (struct sockaddr_in *) &(ifr.ifr_addr);
  int sock;
  FILE *resolv;

#ifdef __SOLARIS
  sock = socket (PF_INET, SOCK_RAW, htons (0));
#else
  sock = socket (AF_PACKET, SOCK_PACKET, htons (ETH_P_ALL));
#endif
  memset (&ifr, '\0', sizeof (struct ifreq));
  memcpy (ifr.ifr_name, iface_name, strlen (iface_name));
  p->sin_family = AF_INET;
  (p->sin_addr).s_addr = (net_conf->myIP).s_addr;
/* setting IP address */
  if (-1 == ioctl (sock, SIOCSIFADDR, &ifr))
    {
      err_msg (MSG_WARN, "ioctl() failed setting IP: %s", strerror (errno));
      close (sock);
      return -1;
    }
/* setting netmask */
  if ((net_conf->netmask).s_addr != INADDR_ANY)
    {
      (p->sin_addr).s_addr = (net_conf->netmask).s_addr;
      if (-1 == ioctl (sock, SIOCSIFNETMASK, &ifr))
	{
	  err_msg (MSG_WARN, "ioctl() failed setting netmask: %s",
		   strerror (errno));
	  close (sock);
	  return -1;
	}
    }
/* if we just had to reset the NIC, our job is done, we exit */
  if (!((net_conf->what_is_set) & GW_IP))
    {
      err_msg (DBG_INFO, "interface %s resetted", iface_name);
      return 0;
    }
/* 'route add default gw' on gateway IP
 * first we'll work with the p struct to manage better the data in rtetry 
 * struct, then the ioctl things the swiss knife on linux :-) */
  memset (&rtent, 0, sizeof (struct rtentry));
  p = (struct sockaddr_in *) &rtent.rt_dst;
  p->sin_family = AF_INET;
/* destination */
  p->sin_addr.s_addr = INADDR_ANY;
  p = (struct sockaddr_in *) &rtent.rt_gateway;
  p->sin_family = AF_INET;
/*gateway */
  memcpy (&((p->sin_addr).s_addr), &((net_conf->gatewayIP).s_addr), 4);
#ifndef __SOLARIS
  p = (struct sockaddr_in *) &rtent.rt_genmask;
  p->sin_family = AF_INET;
/* destination */
  p->sin_addr.s_addr = INADDR_ANY;
  rtent.rt_dev = iface_name;
  rtent.rt_metric = 1;
#endif
  rtent.rt_flags = RTF_UP | RTF_GATEWAY;
  if (-1 == ioctl (sock, SIOCADDRT, &rtent))
    {
      err_msg (MSG_WARN, "ioctl() failed setting route: %s",
	       strerror (errno));
      close (sock);
      return -1;
    }
  close (sock);
/* last, if we have DNSs we overwrite /etc/resolv.conf */
  if ((net_conf->what_is_set) & DNS1 || (net_conf->what_is_set) & DNS2)
    {
      if (NULL == (resolv = fopen ("/etc/resolv.conf", "w")))
	{
	  err_msg (MSG_ERR, "error opening /etc/resolv.conf: %s",
		   strerror (errno));
	  err_msg (MSG_WARN, "could not set DNSs");
	}
      else
	{
	  if ((net_conf->what_is_set) & DNS1)
	    fprintf (resolv, "nameserver %s", inet_ntoa (net_conf->dns1));
	  if ((net_conf->what_is_set) & DNS2)
	    fprintf (resolv, "nameserver %s", inet_ntoa (net_conf->dns2));
	  fclose (resolv);
	}
    }
  return 0;
}

/* function netconf_unapply()
 * this function resets to zero any network configuration previously applied
 * to the NIC. It returns 0 on success, -1 on errors.
 */
int
netconf_unapply (char *iface_name)
{
  netconfig_t zero_conf;

  memset (&zero_conf, 0, sizeof (netconfig_t));
//      (zero_conf.gatewayIP).s_addr=INADDR_ANY;
//      (zero_conf.netmask).s_addr=INADDR_ANY;
//      (zero_conf.myIP).s_addr=INADDR_ANY;
  err_msg (MSG_INFO, "resetting the NIC %s", iface_name);
  return netconf_apply (iface_name, &zero_conf);
}

/* function selfconf_read()
 * selfconf_read() is used to read an xml configuration file, which contains
 * commonly used parameters. The first argument is the configuration file name,
 * all the others are the parameters to be set (program_flags is not passed
 * because is a global variable). The function returns 1 if succeeds, -1 if the
 * file does not exist.
 */
int
selfconf_read (char *conf_fname, char *iface, char *mac,
	       char **prof_fname, int *timeout)
{
  xmlDocPtr doc;
  xmlNodePtr cur;
  xmlChar *key;
  int temp;

/* xml file opening */
  if (NULL == (doc = xmlParseFile (conf_fname)))
    {
      if (ENOENT == errno)
	return -1;
      else
	err_quit ("could not open configuration file %s: %s", conf_fname,
		  strerror (errno));
    }
/* we check for the file being not empty */
  if (NULL == (cur = xmlDocGetRootElement (doc)))
    {
      err_msg (MSG_WARN, "empty configuration file %s", conf_fname);
      xmlFreeDoc (doc);
      return -1;
    }
/* we check for the file being of proper type */
  if (xmlStrcmp (cur->name, (const xmlChar *) "selfDHCPConfig"))
    {
      xmlFreeDoc (doc);
      err_quit ("configuration file %s of the wrong type, root node is not"
		" selfDHCPConfig", prof_fname);
    }
  cur = cur->xmlChildrenNode;
  err_msg (DBG_INFO, "reading configuration");
/* while we got children elements of the current configuration, we check one by
 * one the content */
  while (cur != NULL)
    {
/* active mode */
      if ((!xmlStrcmp (cur->name, (const xmlChar *) "activeMode")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= ACTIVE_MODE;
	  err_msg (DBG_INFO, "configuration: active mode");
	  xmlFree (key);
	}
/* arp flooding */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "arpFlooding")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= ARP_POISON_ENA;
	  err_msg (DBG_INFO, "configuration: arp flood");
	  xmlFree (key);
	}
/* console output */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "consoleOutput")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= TEXTONLY_MODE;
	  err_msg (DBG_INFO, "configuration: console output");
	  xmlFree (key);
	}
/* daemon */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "daemonMode")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= IS_DAEMON_PROC;
	  err_msg (DBG_INFO, "configuration: daemon mode");
	  xmlFree (key);
	}
/* debug */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "debugLevel")))
	{
/* any parameters from comman line overrids the one in the configuration; anyway
 * because debug level is the only one we set before reading configuration, we
 * have to skip the configuration if already set. */
	  if (program_flags & DEBUG_MODE)
	    break;
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= DEBUG_MODE;
	  if (NULL != key)
	    temp = atoi ((char *) key);
	  else
	    temp = 1;
	  if (temp < 1 || temp > 4)
	    err_quit ("wrong debug level %d, should be [1-4]", temp);
	  if (temp < (program_flags & DEBUG_LEVEL))
	    program_flags |= (temp - 1);
	  err_msg (DBG_INFO, "configuration: debug level %d", temp);
	  xmlFree (key);
	}
/* iface name */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "interfaceName")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (NULL != key)
	    iface = strdup ((char *) key);
	  else
	    err_quit ("<interfaceName> tag in file %s cannot be void",
		      conf_fname);
	  err_msg (DBG_INFO, "configuration: interface %s", iface);
	  xmlFree (key);
	}
/* mac */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "macAddress")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (NULL != key)
	    mac = strdup ((char *) key);
	  else
	    err_quit ("<macAddress> tag in file %s cannot be void",
		      conf_fname);
	  err_msg (DBG_INFO, "configuration: mac %s", mac);
	  xmlFree (key);
	}
/* profiles file name */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "profilesFile")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (NULL != key)
	    *prof_fname = strdup ((char *) key);
	  program_flags |= PROFILE_ENABLED;
	  err_msg (DBG_INFO, "configuration: profiles file %s", prof_fname);
	  xmlFree (key);
	}
/* reset NIC */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "resetNIC")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= RESET_NIC;
	  err_msg (DBG_INFO, "configuration: reset NIC");
	  xmlFree (key);
	}
/* timeout */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "timeout")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  if (NULL != key)
	    *timeout = atoi ((char *) key);
	  else
	    err_quit ("<timeout> tag in file %s cannot be void", conf_fname);
	  err_msg (DBG_INFO, "configuration: timeout %d", *timeout);
	  xmlFree (key);
	}
/* profiles update */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "updateProfiles")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= PROFILE_UPDATE;
	  err_msg (DBG_INFO, "configuration: profiles update");
	  xmlFree (key);
	}
/* verbose */
      else if ((!xmlStrcmp (cur->name, (const xmlChar *) "verboseMode")))
	{
	  key = xmlNodeListGetString (doc, cur->xmlChildrenNode, 1);
	  err_msg (DBG_PROG, "xml element \"%s\"", (char *) key);
	  program_flags |= VERBOSE_MODE;
	  err_msg (DBG_INFO, "configuration: verbose mode");
	  xmlFree (key);
	}
/* junk */
      else;
/* then we get to the next element */
      cur = cur->next;
    }
  xmlFreeDoc (doc);
  return 1;
}

/* function selfconf_write()
 * selfconf_write() writes the current configuration to the file passed through
 * conf_name.
 */
void
selfconf_write (char *conf_fname, char *iface, char *mac,
		char *prof_fname, int *timeout)
{
  FILE *doc_;

  if (NULL == (doc_ = fopen (conf_fname, "w")))
    err_quit ("fopen() on file %s failed: %s", conf_fname, strerror (errno));
  fprintf (doc_, "<?xml version=\"\1.0\"?>\n");
  fprintf (doc_, "<selfDHCPConfig>\n");
  if (program_flags & ARP_POISON_ENA)
    printf ("  <arpFlooding>true</arpFlooding>\n");
  if (program_flags & TEXTONLY_MODE)
    printf ("  <consoleOutput>true</consoleOutput>\n");
  if (program_flags & IS_DAEMON_PROC)
    printf ("  <daemonMode>true</daemonMode>\n");
  if (program_flags & DEBUG_MODE)
    printf ("  <debugLevel>%d</debugLevel>\n", program_flags & DEBUG_LEVEL);
  printf ("  <interfaceName>%s</interfaceName>\n", iface);
  if (mac != NULL)
    printf ("  <macAddress>%s</macAddress>\n", mac);
  if (program_flags & PROFILE_ENABLED)
    printf ("  <profilesFile>%s</profilesFile>\n", prof_fname);
  if (program_flags & RESET_NIC)
    printf ("  <resetNIC>true</resetNIC>\n");
  printf ("  <timeout>%d</timeout>\n", *timeout);
  if (program_flags & PROFILE_UPDATE)
    printf ("  <updateProfiles>true</updateProfiles>\n");
  if (program_flags & VERBOSE_MODE)
    printf ("  <verboseMode>true</verboseMode>\n");
  fprintf (doc_, "</selfDHCPConfig>\n");
  fclose (doc_);
  return;
}

/* function change_mac() 
 * This function is used to change the MAC address of the local NIC, in case we
 * don't want to use the real one.
 */
void
change_mac (char *mac_addr_str, char *iface_name)
{
#ifdef __SOLARIS
/* in Solaris we haven't found a way of changing the hardware address, so just
 * display an error message */
  err_msg (MSG_ERR, "this function is not implemented under Solaris");
  return;
#else
  u_char mac_addr_num[ETH_ALEN];
/* we have to use two structures, or ioctl() won't recognize the iface after
 * changing the MAC and will give an "Invalid argument" error */
  struct ifreq ifr1, ifr2;
  int sock;
  char buf[25];

  if (-1 == (sock = socket (AF_INET, SOCK_DGRAM, 0)))
    {
      close (sock);
      err_quit ("error opening socket: %s", strerror (errno));
    }
  snprintf (ifr1.ifr_name, IF_NAMESIZE, "%s", iface_name);
/* first, we get the iface current MAC */
  if (-1 == ioctl (sock, SIOCGIFHWADDR, &ifr1))
    {
      close (sock);
      err_quit ("ioctl() failed while reading iface name: %s",
		strerror (errno));
    }
  err_msg (DBG_INFO, "current MAC: %s",
	   mac2str ((u_char *) & (ifr1.ifr_hwaddr.sa_data), buf));
  memcpy (&ifr2, &ifr1, sizeof (struct ifreq));
/* second we bring iface down (else we'd get a "Device busy" error) */
  ifr1.ifr_flags &= (~IFF_UP);
  if (-1 == ioctl (sock, SIOCSIFFLAGS, &ifr1))
    {
      close (sock);
      err_quit ("ioctl() failed while bringing iface down: %s",
		strerror (errno));
    }
/* third, we change the MAC */
  str2mac (mac_addr_str, mac_addr_num);
  memcpy (&(ifr2.ifr_hwaddr.sa_data), mac_addr_num, ETH_ALEN);
  if (-1 == ioctl (sock, SIOCSIFHWADDR, &ifr2))
    {
      close (sock);
      err_quit ("ioctl() failed while changing iface MAC: %s",
		strerror (errno));
    }
  err_msg (MSG_INFO, "MAC changed to %s",
	   mac2str ((u_char *) & (ifr2.ifr_hwaddr.sa_data), buf));
/* last, we bring the iface back up, or libpcap will complain */
  ifr2.ifr_flags &= IFF_UP;
  if (-1 == ioctl (sock, SIOCSIFFLAGS, &ifr2))
    {
      close (sock);
      err_quit ("ioctl() failed while bringing iface up: %s",
		strerror (errno));
    }
  close (sock);
  return;
#endif
}

/* wrapped fgets() */
char *
Fgets (char *s, int size, FILE * stream)
{
  char *result;
  if (NULL == (result = fgets (s, size, stream)))
    err_msg (MSG_ERR, "fgets() failed");
  return result;
}
