/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_pthread.c
 * purposes:	functions for the thread-oriented implementation (including
 * 				wrapped pthread_*() functions)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

/* The error messages within #ifdef DEBUG are to be used for VERY low level
 * debug only (if searching for deadlock or insertion problems) */

#include "selfdhcp_pthread.h"

/* function daemon_init()
 * used to initialise the daemon. Almost as in Stevens UNP.
 */
int
daemon_init (void)
{
  int i;
  pid_t pid;

  if ((pid = fork ()) < 0)
    {
      return -1;
    }
  else if (pid != 0)
    {
      exit (0);			/* parent goes bye-bye */
    }

  /* child continues */
  setsid ();			/* become session leader */
  signal (SIGHUP, SIG_IGN);

  if ((pid = fork ()) < 0)
    {
      return -1;
    }
  else if (pid != 0)
    {
      exit (0);			/* parent goes bye-bye */
    }

  chdir ("/");			/* change working directory */
  umask (0);			/* clear our file mode creation mask */

  for (i = 0; i < MAX_FD; i++)
    close (i);

  openlog (prog_name, LOG_PID, LOG_DAEMON);

  return 0;
}

/* function semaphores_init()
 * used to initialise the semaphores within a given struct. type must be chosen
 * within the appropriate #define.
 */
void
semaphores_init (void *toinit, int type)
{
  int *nreaders, *nwriters, count;
  pthread_mutex_t *sem;
  switch (type)
    {
    case LISTA_T:
      sem = ((lista_t *) toinit)->semafori;
      nreaders = &(((lista_t *) toinit)->numreaders);
      nwriters = &(((lista_t *) toinit)->numwriters);
      break;
    case SUBLIST_T:
      sem = ((sublist_t *) toinit)->semafori;
      nreaders = &(((sublist_t *) toinit)->numreaders);
      nwriters = &(((sublist_t *) toinit)->numwriters);
      break;
    case NETCONF_T:
      sem = ((netconfig_t *) toinit)->semafori;
      nreaders = &(((netconfig_t *) toinit)->numreaders);
      nwriters = &(((netconfig_t *) toinit)->numwriters);
    case IPLIST_T:
      sem = ((ip2test_t *) toinit)->semafori;
      nreaders = &(((ip2test_t *) toinit)->numreaders);
      nwriters = &(((ip2test_t *) toinit)->numwriters);
    default:
      return;
    }
  nreaders = nwriters = 0;
  for (count = SEM_READERS; count <= SEM_MESC2; count++)
    pthread_mutex_init (&sem[count], NULL);
  return;
}

/* function read_reserv()
 * used to reserv one item (passed through tolock) of a list for read. type must
 * be chosen within the appropriate #define. If all was right, 0 is returned,
 * -1 else.
 */
int
read_reserv (void *tolock, int type)
{
  int *nreaders;
  pthread_mutex_t *sem;
  char buffer[50];

  if (tolock == NULL)
    err_quit ("read_reserv() got NULL pointer to lock (probably a bug)");
  switch (type)
    {
    case LISTA_T:
      sem = ((lista_t *) tolock)->semafori;
      nreaders = &(((lista_t *) tolock)->numreaders);
      break;
    case SUBLIST_T:
      sem = ((sublist_t *) tolock)->semafori;
      nreaders = &(((sublist_t *) tolock)->numreaders);
      break;
    case NETCONF_T:
      sem = ((netconfig_t *) tolock)->semafori;
      nreaders = &(((netconfig_t *) tolock)->numreaders);
      break;
    case IPLIST_T:
      sem = ((ip2test_t *) tolock)->semafori;
      nreaders = &(((ip2test_t *) tolock)->numreaders);
      break;
    default:
      return -1;
    }
  err_msg (DBG_LOWLEV, "read_reserv(): trying to lock item @%p"
	   "(currently %d readers)... ", tolock, *nreaders);
  if (0 != pthread_mutex_lock (&sem[SEM_MESC2]) ||
      0 != pthread_mutex_lock (&sem[SEM_READERS]) ||
      0 != pthread_mutex_lock (&sem[SEM_MESC]))
    {
      err_msg (DBG_LOWLEV, "read_reserv(): could not lock semaphores in"
	       " item @%p: %s", tolock, strerror (errno));
      return -1;
    }
  (*nreaders)++;
  if ((*nreaders) < 0)
    err_msg (DBG_LOWLEV, "read_reserv(%p): %d readers", tolock, *nreaders);
  if (*nreaders == 1)
    if (0 != pthread_mutex_lock (&sem[SEM_WRITERS]))
      {
	err_msg (DBG_LOWLEV, "read_reserv(): could not lock semaphore"
		 " WRITERS: %s", strerror (errno));
	return -1;
      }
  if (0 != pthread_mutex_unlock (&sem[SEM_MESC]) ||
      0 != pthread_mutex_unlock (&sem[SEM_READERS]) ||
      0 != pthread_mutex_unlock (&sem[SEM_MESC2]))
    {
      err_msg (DBG_LOWLEV, "read_reserv(): could not unlock semaphores: %s",
	       strerror (errno));
      return -1;
    }
  err_msg (DBG_LOWLEV,
	   "read_reserv() :item @%p locked (currently %d readers)", tolock,
	   *nreaders);
  return 0;
}

/* function read_release()
 * used to release one item (passed through tounlock) of a list previously
 * locked through read_reserv(). Type must be chosen within the appropriate
 * #define. If all was right, 0 is returned,-1 else.
 */
int
read_release (void *tounlock, int type)
{
  int *nreaders;
  pthread_mutex_t *sem;
  if (tounlock == NULL)
    err_quit ("read_release() got NULL pointer to unlock (probably a bug)");
  switch (type)
    {
    case LISTA_T:
      sem = ((lista_t *) tounlock)->semafori;
      nreaders = &(((lista_t *) tounlock)->numreaders);
      break;
    case SUBLIST_T:
      sem = ((sublist_t *) tounlock)->semafori;
      nreaders = &(((sublist_t *) tounlock)->numreaders);
      break;
    case NETCONF_T:
      sem = ((netconfig_t *) tounlock)->semafori;
      nreaders = &(((netconfig_t *) tounlock)->numreaders);
      break;
    case IPLIST_T:
      sem = ((ip2test_t *) tounlock)->semafori;
      nreaders = &(((ip2test_t *) tounlock)->numreaders);
      break;
    default:
      return -1;
    }
  err_msg (DBG_LOWLEV, "read_release(): trying to unlock item @%p"
	   " (currently %d readers)... ", tounlock, *nreaders);
  if (0 != pthread_mutex_lock (&sem[SEM_MESC]))
    {
      err_msg (DBG_LOWLEV, "read_release(): could not lock semaphore MESC:"
	       " %s", strerror (errno));
      return -1;
    }
  (*nreaders)--;
  if ((*nreaders) < 0)
    err_msg (DBG_LOWLEV, "read_release(%p): %d readers", tounlock, *nreaders);
  if (*nreaders == 0)
    if (0 != pthread_mutex_unlock (&sem[SEM_WRITERS]))
      {
	err_msg (DBG_LOWLEV, "read_release(): could not unlock semaphore"
		 " WRITERS: %s", strerror (errno));
	return -1;
      }
  if (0 != pthread_mutex_unlock (&sem[SEM_MESC]))
    {
      err_msg (DBG_LOWLEV, "read_release(): could not unlock semaphore MESC:"
	       " %s", strerror (errno));
      return -1;
    }
  err_msg (DBG_LOWLEV, "read_release(): item @%p unlocked"
	   " (currently %d readers)", tounlock, *nreaders);
  return 0;
}

/* function write_reserv()
 * used to reserv one item (passed through tolock) of a list for write. Type
 * must be chosen within the appropriate #define. If all was right, 0 is
 * returned, -1 else.
 */
int
write_reserv (void *tolock, int type)
{
  int *nwriters;
  pthread_mutex_t *sem;
  if (tolock == NULL)
    err_quit ("write_reserv() got NULL pointer to lock (probably a bug)");
  switch (type)
    {
    case LISTA_T:
      sem = ((lista_t *) tolock)->semafori;
      nwriters = &(((lista_t *) tolock)->numwriters);
      break;
    case SUBLIST_T:
      sem = ((sublist_t *) tolock)->semafori;
      nwriters = &(((sublist_t *) tolock)->numwriters);
      break;
    case NETCONF_T:
      sem = ((netconfig_t *) tolock)->semafori;
      nwriters = &(((netconfig_t *) tolock)->numwriters);
      break;
    case IPLIST_T:
      sem = ((ip2test_t *) tolock)->semafori;
      nwriters = &(((ip2test_t *) tolock)->numwriters);
      break;
    default:
      return -1;
    }
  err_msg (DBG_LOWLEV, "write_reserv(): trying to lock item @%p"
	   " (currently %d writers)... ", tolock, *nwriters);
  if (0 != pthread_mutex_lock (&sem[SEM_MESC1]))
    {
      err_msg (DBG_LOWLEV, "write_reserv(): could not lock semaphor MESC1"
	       " %s", strerror (errno));
      return -1;
    }
  (*nwriters)++;
  if ((*nwriters) < 0)
    {
      err_quit ("write_reserv(%p): %d writers", tolock, *nwriters);
    }
  if ((*nwriters) == 1)
    if (0 != pthread_mutex_lock (&sem[SEM_READERS]))
      {
	err_msg (DBG_LOWLEV, "write_reserv(): could not lock semaphor READERS"
		 ": %s", strerror (errno));
	return -1;
      }
  if (0 != pthread_mutex_unlock (&sem[SEM_MESC1]))
    {
      err_msg (DBG_LOWLEV, "write_reserv(): could not unlock semaphore MESC1:"
	       " %s", strerror (errno));
      return -1;
    }
  if (0 != pthread_mutex_lock (&sem[SEM_WRITERS]))
    {
      err_msg (DBG_LOWLEV, "write_reserv(): could not lock semaphore WRITERS"
	       ": %s", strerror (errno));
      return -1;
    }
  err_msg (DBG_LOWLEV,
	   "write_reserv(): item @%p locked (currently %d writers)", tolock,
	   *nwriters);
  return 0;
}

/* function write_release()
 * used to release one item (passed through tounlock) of a list previously
 * locked through write_reserv(). Type must be chosen within the appropriate
 * #define. If all was right, 0 is returned,-1 else.
 */
int
write_release (void *tounlock, int type)
{
  int *nwriters;
  pthread_mutex_t *sem;
  if (tounlock == NULL)
    err_quit ("write_release() got NULL pointer to unlock (probably a bug)");
  switch (type)
    {
    case LISTA_T:
      sem = ((lista_t *) tounlock)->semafori;
      nwriters = &(((lista_t *) tounlock)->numwriters);
      break;
    case SUBLIST_T:
      sem = ((sublist_t *) tounlock)->semafori;
      nwriters = &(((sublist_t *) tounlock)->numwriters);
      break;
    case NETCONF_T:
      sem = ((netconfig_t *) tounlock)->semafori;
      nwriters = &(((netconfig_t *) tounlock)->numwriters);
      break;
    case IPLIST_T:
      sem = ((ip2test_t *) tounlock)->semafori;
      nwriters = &(((ip2test_t *) tounlock)->numwriters);
      break;
    default:
      return -1;
    }
  err_msg (DBG_LOWLEV, "write_release(): trying to unlock item @%p"
	   " (currently %d writers)... ", tounlock, *nwriters);
  if (0 != pthread_mutex_unlock (&sem[SEM_WRITERS]))
    {
      err_msg (DBG_LOWLEV,
	       "write_release(): could not unlock semaphore WRITERS" ": %s",
	       strerror (errno));
      return -1;
    }
  if (0 != pthread_mutex_lock (&sem[SEM_MESC1]))
    {
      err_msg (DBG_LOWLEV, "write_release(): could not lock semaphore MESC1:"
	       " %s", strerror (errno));
      return -1;
    }
  (*nwriters)--;
  if ((*nwriters) < 0)
    err_quit ("write_release(%p): %d writers", tounlock, *nwriters);
  if ((*nwriters) == 0)
    if (0 != pthread_mutex_unlock (&sem[SEM_READERS]))
      {
	err_msg (DBG_LOWLEV, "write_release(): could not unlock semaphore"
		 " READERS: %s", strerror (errno));
	return -1;
      }
  if (0 != pthread_mutex_unlock (&sem[SEM_MESC1]))
    {
      err_msg (DBG_LOWLEV,
	       "write_release(): could not unlock semaphore MESC1:" " %s",
	       strerror (errno));
      return -1;
    }
  err_msg (DBG_LOWLEV, "write_release(): item @%p unlocked "
	   "(currently %d writers)", tounlock, *nwriters);
  return 0;
}

/* wrapped pthread_mutex_lock() */
int
Pthread_mutex_lock (pthread_mutex_t * __mutex)
{
  int retval;

  retval = pthread_mutex_lock (__mutex);
  switch (retval)
    {
    case 0:
      break;
    case EINVAL:
      err_msg (DBG_LOWLEV, "pthread_mutex_lock(): mutex have not been "
	       "properly initialized");
      break;
    case EDEADLK:
      err_msg (DBG_LOWLEV, "pthread_mutex_lock(): mutex already locked "
	       "by the calling thread (probable deadlock)");
      break;
    default:
      err_msg (DBG_LOWLEV, "pthread_mutex_lock(): unknown error locking "
	       "mutex");
      break;
    }
  return retval;
}

/* wrapped pthread_mutex_unlock() */
int
Pthread_mutex_unlock (pthread_mutex_t * __mutex)
{
  int retval;

  retval = pthread_mutex_lock (__mutex);
  switch (retval)
    {
    case 0:
      break;
    case EINVAL:
      err_msg (DBG_LOWLEV, "pthread_mutex_unlock(): mutex have not been "
	       "properly initialized");
      break;
    case EPERM:
      err_msg (DBG_LOWLEV, "pthread_mutex_unlock(): the calling thread "
	       "does not own the mutex");
      break;
    default:
      err_msg (DBG_LOWLEV, "pthread_mutex_unlock(): unknown error "
	       "locking mutex");
      break;
    }
  return retval;
}

/* wrapped pthread_create() */
int
Pthread_create (pthread_t * thread, pthread_attr_t * attr,
		void *(*start_routine) (void *), void *arg)
{
  int retval;

  err_msg (DBG_LIB, "pthread_create(): entering");
  retval = pthread_create (thread, attr, start_routine, arg);
  switch (retval)
    {
    case 0:
      err_msg (DBG_LIB, "pthread_create() succeded");
      break;
    case EAGAIN:
      err_quit ("pthread_create() failed: not enough system "
		"resources to create a process for the new thread");
      break;
    default:
      err_quit ("pthread_create() failed: unknown error");
      break;
    }
  return retval;
}
