/* selfDHCP
 * Copyright (C) 2002-2003 Enrico "WiseLYNX" Manzini
 *
 * module:		selfdhcp_struct.c
 * purposes:	functions for handling the main data structures
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.
 */

/* The error messages within #ifdef DEBUG are to be used for VERY low level
 * debug only (if searching for deadlock or insertion problems) */

#include "selfdhcp_struct.h"

/* function sublistinsert()
 * this function is used to insert an item (MAC for IP, and vice-versa), passed
 * through the reldata argumnet, in the "relatives" sublist of the item element. 
 * A pointer to the created or updated record is returned, or NULL in case of
 * error.
 */
sublist_t *
sublistinsert (lista_t * element, lista_t * reldata)
{
  sublist_t *newEntry, *cursor, *backcursor;
  int nE_howmany;
  err_msg (DBG_LIB, "sublistinsert(): entering");
/* if the relative data is not found, we have to create a record in the list */
  read_reserv (element, LISTA_T);
  if (NULL == (newEntry = sublistsearch (element, reldata)))
    {
      err_msg (DBG_LIB, "sublistinsert(): item @%p not found in sublist",
	       reldata);
      read_release (element, LISTA_T);
      if (NULL == (newEntry = calloc (1, sizeof (sublist_t))))
	{
	  err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
	  return NULL;
	}
      newEntry->item = reldata;
      semaphores_init (newEntry, SUBLIST_T);
      (newEntry->howmany)++;
      (element->relnum)++;
/* if we have no other sublist entries, we link the new one directly to its item */
      read_reserv (element, LISTA_T);
      cursor = element->relatives;
      read_release (element, LISTA_T);
      if (cursor == NULL)
	{
	  write_reserv (element, LISTA_T);
	  element->relatives = newEntry;
	  write_release (element, LISTA_T);
	}
/* else we link into the last position */
      else
	{
	  write_reserv (cursor, SUBLIST_T);
	  while (cursor->nextnum != NULL)
	    {
	      backcursor = cursor;
	      cursor = cursor->nextnum;
	      write_reserv (cursor, SUBLIST_T);
	      write_release (backcursor, SUBLIST_T);
	    }
	  cursor->nextnum = newEntry;
	  write_release (cursor, SUBLIST_T);
	}
    }
/* If sublist entry was already there... */
  else
    {
      err_msg (DBG_LIB, "sublistinsert(): item found in sublist @%p",
	       newEntry);
      write_reserv (newEntry, SUBLIST_T);
/* we update the howmany counter for the sublist element */
      (newEntry->howmany)++;
/* the we have to unlink the element from the sublist */
      if (newEntry != element->relatives)
	{
	  err_msg (DBG_LIB, "sublistinsert(): unlinking item @%p", newEntry);
	  write_release (newEntry, SUBLIST_T);
	  cursor = element->relatives;
	  read_release (element, LISTA_T);
	  write_reserv (cursor, SUBLIST_T);
	  while (cursor->nextnum != newEntry)
	    {
	      backcursor = cursor;
	      cursor = cursor->nextnum;
	      write_reserv (cursor, SUBLIST_T);
	      write_release (backcursor, SUBLIST_T);
	    }
	  cursor->nextnum = newEntry->nextnum;
	  write_release (cursor, SUBLIST_T);
	  newEntry->nextnum = NULL;
	}
      else
	{
	  err_msg (DBG_LIB, "sublistinsert(): no need to unlink item @%p",
		   newEntry);
	  read_release (element, LISTA_T);
	  write_release (newEntry, SUBLIST_T);
	}
/* then we take care of the howmany ordering */
      err_msg (DBG_LIB, "sublistinsert(): taking care of howmany ordering");
      read_reserv (element, LISTA_T);
      cursor = element->relatives;
      read_release (element, LISTA_T);
      if (cursor != NULL)
	write_reserv (cursor, SUBLIST_T);
/* we could already be in head of the sublist... */
      if (newEntry != cursor)
	read_reserv (newEntry, SUBLIST_T);
      if (newEntry == cursor)
	{
	  err_msg (DBG_LIB, "sublistinsert(): head of the list");
	  write_release (cursor, SUBLIST_T);
	}
/* ... or we could became head of the list... */
      else if (cursor != NULL && (newEntry->howmany >= cursor->howmany))
	{
	  err_msg (DBG_LIB, "sublistinsert(): becoming head of the list");
	  read_release (newEntry, SUBLIST_T);
	  newEntry->nextnum = cursor;
	  write_reserv (element, LISTA_T);
	  element->relatives = newEntry;
	  write_release (element, LISTA_T);
	  write_release (cursor, SUBLIST_T);
	}
/* ... or we could be in any other position */
      else
	{
	  err_msg (DBG_LIB, "sublistinsert(): not in head of the list");
	  nE_howmany = newEntry->howmany;
	  read_release (newEntry, SUBLIST_T);
	  while (cursor != NULL && nE_howmany < cursor->howmany)
	    {
	      backcursor = cursor;
	      cursor = cursor->nextnum;
	      if (cursor != NULL)
		write_reserv (cursor, SUBLIST_T);
	      write_release (backcursor, SUBLIST_T);
	    }
	  write_reserv (backcursor, SUBLIST_T);
	  write_reserv (newEntry, SUBLIST_T);
	  newEntry->nextnum = cursor;
	  backcursor->nextnum = newEntry;
	  write_release (newEntry, SUBLIST_T);
	  write_release (backcursor, SUBLIST_T);
	  if (cursor != NULL)
	    write_release (cursor, SUBLIST_T);
	}
    }
  err_msg (DBG_LIB, "sublistinsert(): exiting");
  return newEntry;
}

/* function sublistsearch()
 * searches for a given item reldata in the sublist of the item element.
 * It is returned a pointer to the item found, or NULL if the item was not there.
 */
sublist_t *
sublistsearch (lista_t * element, lista_t * reldata)
{
  sublist_t *cursor, *backcursor;

  err_msg (DBG_LIB, "sublistsearch(): entering");
  err_msg (DBG_LIB, "sublistsearch(): searching item @%p", reldata);
  read_reserv (element, LISTA_T);
  cursor = element->relatives;
  read_release (element, LISTA_T);
  err_msg (DBG_LIB, "sublistsearch(): starting from item @%p", cursor);
  if (cursor == NULL)
    {
      err_msg (DBG_LIB, "sublistsearch(): head of the list");
      return NULL;
    }
  read_reserv (cursor, SUBLIST_T);
  while (cursor != NULL && cursor->item != reldata)
    {
      err_msg (DBG_LIB, "sublistsearch(): scrolling item @%p", cursor);
      backcursor = cursor;
      cursor = cursor->nextnum;
      err_msg (DBG_LIB, "sublistsearch(): next item @%p.", cursor);
      if (cursor != NULL)
	read_reserv (cursor, SUBLIST_T);
      read_release (backcursor, SUBLIST_T);
    }
  if (cursor != NULL)
    read_release (cursor, SUBLIST_T);
  err_msg (DBG_LIB, "sublistsearch(): exiting");
  return cursor;
}

/* function relsort()
 * handle the relnum ordering in the list passed through head, after the
 * insertion of the new item element.
 */
void
relsort (lista_t * head, lista_t * element)
{
  lista_t *cursor, *backcursor;

  err_msg (DBG_LIB, "relsort(): entering");
  write_reserv (head, LISTA_T);
  err_msg (DBG_LIB, "relsort(): head is @%p", head);
  cursor = head->nextrel;
  if (cursor == NULL)
    {
      err_msg (DBG_LIB, "relsort(): first record in");
      err_msg (DBG_LIB, "relsort(): exiting");
      head->nextrel = element;
      write_release (head, LISTA_T);
      return;
    }
  else
    write_reserv (cursor, LISTA_T);
/* Case 1: the elemen is already in the head position */
  if (element == cursor)
    {
      err_msg (DBG_LIB, "relsort(): already in head of list");
      write_release (head, LISTA_T);
    }
  else
    {
/* first we find and unlink element */
      while ((cursor->nextrel != NULL) && (cursor->nextrel != element))
	{
	  err_msg (DBG_LIB,
		   "relsort(): scrolling item @%p (next @%p) for unlinking",
		   cursor, cursor->nextrel);
	  write_reserv (cursor->nextrel, LISTA_T);
	  backcursor = cursor;
	  cursor = cursor->nextrel;
	  write_release (backcursor, LISTA_T);
	}
      if (cursor->nextrel != NULL)
	cursor->nextrel = element->nextrel;
      write_release (cursor, LISTA_T);
/* after unlinking, restart from head */
      cursor = head->nextrel;
      backcursor = head;
      write_reserv (cursor, LISTA_T);
/* we find the insertion point */
      while ((cursor != NULL) && (element->relnum < cursor->relnum))
	{
	  err_msg (DBG_LIB,
		   "relsort(): scrolling item @%p (next @%p) for relinking",
		   cursor, cursor->nextrel);
	  backcursor = cursor;
	  cursor = cursor->nextrel;
	  if (cursor != NULL)
	    write_reserv (cursor, LISTA_T);
	  write_release (backcursor, LISTA_T);
	}
      err_msg (DBG_LIB, "relsort(): stopped @%p, inserting record", cursor);
      if (backcursor != head)
	write_reserv (backcursor, LISTA_T);
      element->nextrel = cursor;
      backcursor->nextrel = element;
      write_release (backcursor, LISTA_T);
      if (backcursor != head)
	write_release (head, LISTA_T);
    }
  if (cursor != NULL)
    write_release (cursor, LISTA_T);
  err_msg (DBG_LIB, "relsort(): exiting");
  return;
}

/* function insert()
 * this function inserts a generic element in the given list. List type is
 * detected through the internal flag datatype.
 * insert() accepts a pointer to the list head, and a void pointer a previously
 * created chunk of data (the function does not check for the real nature of
 * data, so be careful). A pointer to the newly created (or updated) element is
 * returned in case of success, NULL if something was wrong.
 */
lista_t *
insert (lista_t * head, void *newdata)
{
  lista_t *cursor, *backcursor, *newEntry;
//      struct in_addr *newIP, *IPcursor;
//      unsigned char *newMAC, *MACcursor;
  char buffer[50];

  err_msg (DBG_LIB, "insert(): entering");
  read_reserv (head, LISTA_T);
  cursor = head->next;
  read_release (head, LISTA_T);
  switch (head->datatype)
    {
    case IP_TYPE:
      strcpy (buffer, inet_ntoa (*((struct in_addr *) newdata)));
      break;
    case MAC_TYPE:
      mac2str ((unsigned char *) newdata, buffer);
      break;
    }
  err_msg (DBG_LIB, "insert(): gonna insert %s", buffer);
/* Special case: this is the first record */
  if (cursor == NULL)
    {
      if (NULL == (newEntry = calloc (1, sizeof (lista_t))))
	{
	  err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
	  return NULL;
	}
      err_msg (DBG_LIB, "insert(): head of the list @%p", cursor);
      newEntry->data = newdata;
      newEntry->datatype = head->datatype;
      switch (head->datatype)
	{
	case IP_TYPE:
	  strcpy (buffer, inet_ntoa (*((struct in_addr *) (newEntry->data))));
	  break;
	case MAC_TYPE:
	  mac2str ((unsigned char *) (newEntry->data), buffer);
	  break;
	}
      err_msg (DBG_LIB, "insert(): put new record %s @%p", buffer, newEntry);
      newEntry->howmany++;
      semaphores_init (newEntry, LISTA_T);
      write_reserv (head, LISTA_T);
      head->next = newEntry;
      head->nextnum = newEntry;
      (head->howmany) += 1;
      write_release (head, LISTA_T);
      return newEntry;
    }
/* We search through the list for the insertion point */
  newEntry = search (head, newdata, FINDITEM_MODE);
/* If this chunk of data was not present, we have to create a new entry */
  if (newEntry == NULL)
    {
      err_msg (DBG_LIB, "insert(): item not found in list");
      if (NULL == (newEntry = calloc (sizeof (lista_t), 1)))
	{
	  err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
	  return NULL;
	}
      err_msg (DBG_LIB, "insert(): new item created @%p", newEntry);
      newEntry->data = newdata;
      newEntry->datatype = head->datatype;
      semaphores_init (newEntry, LISTA_T);
      write_reserv (newEntry, LISTA_T);
/* If the entry is newly created, it should be placed at the right point of
 * the list */
      cursor = search (head, newdata, INSPOINT_MODE);
      err_msg (DBG_LIB, "insert(): insertion point found @%p", cursor);
      if (cursor != NULL)
	{
	  write_reserv (cursor, LISTA_T);
	  newEntry->next = cursor->next;
	}
/* If insertion point was not found, item has to be inserted in tail */
      else
	{
	  cursor = head->next;
	  read_reserv (cursor, LISTA_T);
	  while (cursor->next != NULL)
	    {
	      backcursor = cursor;
	      cursor = cursor->next;
	      read_reserv (cursor, LISTA_T);
	      read_release (backcursor, LISTA_T);
	    }
	  read_release (cursor, LISTA_T);
	  write_reserv (cursor, LISTA_T);
	}
      cursor->next = newEntry;
      write_release (cursor, LISTA_T);
/* we update the total number of addresses seen */
      write_reserv (head, LISTA_T);
      (head->howmany) += 1;
      write_release (head, LISTA_T);
      err_msg (DBG_LIB, "insert(): prev: @%p  next: @%p", cursor,
	       newEntry->next);
    }
/* If the data was already present, we just update the count and unlink from the
 * count chain */
  else
    {
      err_msg (DBG_LIB, "insert(): item found @%p", newEntry);
      write_reserv (newEntry, LISTA_T);
      read_reserv (head, LISTA_T);
      if (newEntry != head->nextnum)
	{
	  cursor = head->nextnum;
	  read_release (head, LISTA_T);
	  write_reserv (cursor, LISTA_T);
	  while (cursor->nextnum != newEntry)
	    {
	      backcursor = cursor;
	      cursor = cursor->nextnum;
	      write_reserv (cursor, LISTA_T);
	      write_release (backcursor, LISTA_T);
	    }
	  cursor->nextnum = newEntry->nextnum;
	  write_release (cursor, LISTA_T);
	  newEntry->nextnum = NULL;
	}
      else
	read_release (head, LISTA_T);
    }
/* We take care of the 'howmany' ordering */
  (newEntry->howmany)++;
  write_release (newEntry, LISTA_T);
  err_msg (DBG_LIB, "insert(): this element howmany is %d",
	   newEntry->howmany);
  read_reserv (head, LISTA_T);
  cursor = head->nextnum;
  write_reserv (cursor, LISTA_T);
  read_release (head, LISTA_T);
/* Case 1: the elemen is already in the head position */
  if (newEntry == cursor)
    err_msg (DBG_LIB, "insert(): already in head for nextnum");
/* Case 2: the element becomes the head */
  else if (newEntry->howmany >= cursor->howmany)
    {
      err_msg (DBG_LIB, "insert(): inserting on list head for nextnum");
      newEntry->nextnum = cursor;
      write_reserv (head, LISTA_T);
      head->nextnum = newEntry;
      write_release (head, LISTA_T);
    }
/* Case 3: anything else */
  else
    {
      while ((cursor != NULL) && (newEntry->howmany < cursor->howmany))
	{
	  err_msg (DBG_LIB, "insert(): scrolling entry at %p for nextnum",
		   cursor);
	  backcursor = cursor;
	  cursor = cursor->nextnum;
	  if (cursor != NULL)
	    write_reserv (cursor, LISTA_T);
	  write_release (backcursor, LISTA_T);
	}
      err_msg (DBG_LIB, "insert(): stopped scrolling for nextnum at item %p",
	       cursor);
      err_msg (DBG_LIB, "insert(): inserting for nextnum before item at %p",
	       cursor);
      write_reserv (backcursor, LISTA_T);
      newEntry->nextnum = cursor;
      backcursor->nextnum = newEntry;
      write_release (backcursor, LISTA_T);
    }
  if (cursor != NULL)
    write_release (cursor, LISTA_T);
  err_msg (DBG_LIB, "insert(): inserted at address %p (->nextnum @%p)",
	   newEntry, newEntry->nextnum);
  err_msg (DBG_LIB, "insert(): exiting");
  return newEntry;
}


/* function search()
 * this function searches in the list pointed from *head for the data passed
 * through the void *data (data type is recognized through the datatype flag)
 * from the list, but no check is done on the data itself).
 * search() returns a pointer to the element containing the given data, or NULL 
 * if could not find it.
 */
lista_t *
search (lista_t * head, void *searchdata, int mode)
{
  char buffer[30];
  lista_t *cursor, *backcursor;
  int type;
  struct in_addr *searchIP, *curIP;
  unsigned char *searchMAC, *curMAC;

  err_msg (DBG_LIB, "search(): entering");
  read_reserv (head, LISTA_T);
  type = head->datatype;
  cursor = head->next;
  if (cursor != NULL)
    read_reserv (cursor, LISTA_T);
/* if in INSPOINT_MODE we have to check that the head itself is not the
 * insertion point we're looking for */
  if (mode == INSPOINT_MODE && cursor != NULL)
    {
      switch (type)
	{
	case IP_TYPE:
	  searchIP = (struct in_addr *) searchdata;
	  curIP = (struct in_addr *) (cursor->data);
	  if (0 > ipcmp (searchIP, curIP))
	    {
	      read_release (head, LISTA_T);
	      read_release (cursor, LISTA_T);
	      return head;
	    }
	case MAC_TYPE:
	  searchMAC = (unsigned char *) searchdata;
	  curMAC = (unsigned char *) (cursor->data);
	  if (0 > maccmp (searchMAC, curMAC))
	    {
	      read_release (head, LISTA_T);
	      read_release (cursor, LISTA_T);
	      return head;
	    }
	}
    }
  if (cursor != NULL)
    read_release (cursor, LISTA_T);
  read_release (head, LISTA_T);
  err_msg (DBG_LIB, "search(): starting from head @%p", cursor);
  switch (type)
    {
    case IP_TYPE:
      searchIP = (struct in_addr *) searchdata;
      if (cursor != NULL)
	{
	  read_reserv (cursor, LISTA_T);
	  curIP = (struct in_addr *) (cursor->data);
	}
      err_msg (DBG_LIB, "search(): searching IP %s", inet_ntoa (*searchIP));
      while (cursor != NULL)
	{
	  err_msg (DBG_LIB, "search(): scrolling IP %s @%p (->next %p)",
		   inet_ntoa (*curIP), cursor, cursor->next);
	  curIP = (struct in_addr *) (cursor->data);
/* in FINDITEM_MODE we stop if searched IP is found exactly */
	  if (mode == FINDITEM_MODE && 0 == ipcmp (searchIP, curIP))
	    break;
/* in INSPOINT_MODE we stop right after the immediately lesser IP address we
 * find */
	  if (mode == INSPOINT_MODE && 0 > ipcmp (searchIP, curIP))
	    break;
	  if (cursor->next != NULL)
	    read_reserv (cursor->next, LISTA_T);
	  backcursor = cursor;
	  cursor = cursor->next;
	  read_release (backcursor, LISTA_T);
	}
      break;
    case MAC_TYPE:
      searchMAC = (unsigned char *) searchdata;
      if (cursor != NULL)
	{
	  read_reserv (cursor, LISTA_T);
	  curMAC = (unsigned char *) (cursor->data);
	}
      err_msg (DBG_LIB, "search(): searching MAC %s",
	       mac2str (searchMAC, buffer));
      while (cursor != NULL)
	{
	  err_msg (DBG_LIB, "search(): scrolling MAC %s @%p (->next %p)",
		   mac2str (curMAC, buffer), cursor, cursor->next);
	  curMAC = (unsigned char *) (cursor->data);
/* in FINDITEM_MODE we stop if searched MAC is found exactly */
	  if (mode == FINDITEM_MODE && 0 == maccmp (searchMAC, curMAC))
	    break;
/* in INSPOINT_MODE we stop right after the immediately lesser MAC address we
 * find */
	  if (mode == INSPOINT_MODE && 0 > maccmp (searchMAC, curMAC))
	    break;
	  if (cursor->next != NULL)
	    read_reserv (cursor->next, LISTA_T);
	  backcursor = cursor;
	  cursor = cursor->next;
	  read_release (backcursor, LISTA_T);
	}
      break;
    }
  err_msg (DBG_LIB, "search(): found at address %p", cursor);
  if (cursor != NULL)
    read_release (cursor, LISTA_T);
  err_msg (DBG_LIB, "search(): exiting");
  return cursor;
}

/* function findlevel()
 * findlevel returns the first element of nth 'level' (being level the position)
 * in the count list, with elements with the same counter on the same level).
 * NULL is returned if the end of the list is reached without finding elements
 * of the required level.
 */
lista_t *
findlevel (lista_t * head, int level, int mode)
{
  unsigned long num;
  lista_t *cursor, *backcursor;
/* if in LAST_ONE mode, we have to find which the last level is */
  if (level == LAST_ONE)
    {
      level = 1;
      switch (mode)
	{
	case HOW_MANY_TIMES:
	  while (NULL != findlevel (head, level, HOW_MANY_TIMES))
	    level++;
	  break;
	case HOW_MANY_RELATIVES:
	  while (NULL != findlevel (head, level, HOW_MANY_RELATIVES))
	    level++;
	  break;
	}
      level--;
    }
  read_reserv (head, LISTA_T);
  switch (mode)
    {
    case HOW_MANY_TIMES:
      cursor = head->nextnum;
      break;
    case HOW_MANY_RELATIVES:
      cursor = head->nextrel;
      break;
    default:
      err_msg (MSG_ERR, "findlevel(): invalid argument mode=%d"
	       " (probably a bug)", mode);
      return NULL;
    }
  if (cursor != NULL)
    {
      read_reserv (cursor, LISTA_T);
    }
  read_release (head, LISTA_T);
/* main loop */
  while (level > 1 && cursor != NULL)
    {
      backcursor = cursor;
      switch (mode)
	{
	case HOW_MANY_TIMES:
	  if (cursor->nextnum == NULL)
	    return NULL;
	  read_reserv (cursor->nextnum, LISTA_T);
	  cursor = cursor->nextnum;
	  read_release (backcursor, LISTA_T);
	  num = cursor->howmany;
	  while (cursor != NULL && (cursor->howmany) == num)
	    {
	      backcursor = cursor;
	      if (cursor->nextnum != NULL)
		{
		  read_reserv (cursor->nextnum, LISTA_T);
		}
	      cursor = cursor->nextnum;
	      read_release (backcursor, LISTA_T);
	    }
	  level--;
	  break;
	case HOW_MANY_RELATIVES:
	  if (cursor->nextrel == NULL)
	    return NULL;
	  read_reserv (cursor->nextrel, LISTA_T);
	  cursor = cursor->nextrel;
	  read_release (backcursor, LISTA_T);
	  num = cursor->relnum;
	  while (cursor != NULL && (cursor->relnum) == num)
	    {
	      backcursor = cursor;
	      if (cursor->nextrel != NULL)
		{
		  read_reserv (cursor->nextrel, LISTA_T);
		}
	      cursor = cursor->nextrel;
	      read_release (backcursor, LISTA_T);
	    }
	  level--;
	  break;
	}
    }
  if (cursor != NULL)
    read_release (cursor, LISTA_T);
  return cursor;
}

/* function findnext()
 * Given an element in a list, findnext() returns the next element with on the
 * same level, or NULL if there are no more elements on that level.
 */
lista_t *
findnext (lista_t * element, int mode)
{
  lista_t *nextelement;
  unsigned long num;
  switch (mode)
    {
    case HOW_MANY_TIMES:
      read_reserv (element, LISTA_T);
      num = element->howmany;
      nextelement = element->nextnum;
      if (nextelement != NULL)
	read_reserv (nextelement, LISTA_T);
      read_release (element, LISTA_T);
      if (nextelement != NULL && (nextelement->howmany) < num)
	{
	  read_release (nextelement, LISTA_T);
	  return NULL;
	}
      break;
    case HOW_MANY_RELATIVES:
      read_reserv (element, LISTA_T);
      num = element->relnum;
      nextelement = element->nextrel;
      if (nextelement != NULL)
	read_reserv (nextelement, LISTA_T);
      read_release (element, LISTA_T);
      if (nextelement != NULL && (nextelement->relnum) < num)
	{
	  read_release (nextelement, LISTA_T);
	  return NULL;
	}
      break;
    }
  if (nextelement != NULL)
    read_release (nextelement, LISTA_T);
  return nextelement;
}

/* function mac2str()
 * given a MAC address stored as an array of bytes, it returns the string in
 * the format "xx:xx:xx:xx:xx:xx", or NULL if something was wrong.
 */
char *
mac2str (unsigned char *MACaddr_num, char *MACaddr_string)
{
  if (17 != sprintf (MACaddr_string, "%02X:%02X:%02X:%02X:%02X:%02X",
		     MACaddr_num[0],
		     MACaddr_num[1],
		     MACaddr_num[2],
		     MACaddr_num[3], MACaddr_num[4], MACaddr_num[5]))
    return NULL;
  return MACaddr_string;
}

/* function str2mac()
 * given a strig containing a MAC addr in the "xx:xx:xx:xx:xx:xx" format,
 * returns an unsigned char array containing the numeric form of the address,
 * or NULL if something goes wrong.
 */
unsigned char *
str2mac (char *MACaddr_string, unsigned char *MACaddr_num)
{
  if (6 != sscanf (MACaddr_string, "%2x:%2x:%2x:%2x:%2x:%2x",
		   &MACaddr_num[0],
		   &MACaddr_num[1],
		   &MACaddr_num[2],
		   &MACaddr_num[3], &MACaddr_num[4], &MACaddr_num[5]))
    return NULL;
  return &MACaddr_num[0];
}

/* function ipcmp()
 * given two in_addr structures ip1 and ip2, it returns -1, 0 or 1 if ip1 is
 * found respectively lesser than, equal or greater than ip2.
 */
int
ipcmp (struct in_addr *ip1, struct in_addr *ip2)
{
  if (ntohl (ip1->s_addr) < ntohl (ip2->s_addr))
    return -1;
  else if (ntohl (ip1->s_addr) == ntohl (ip2->s_addr))
    return 0;
  else
    return 1;
}

/* function maccmp()
 * given two mac addresses mac1 and mac2, it returns an int lesser than, equal
 * to or greater than zero if mac1 is found respectively lesser than, equal to
 * or greater than mac2.
 */
int
maccmp (unsigned char *mac1, unsigned char *mac2)
{
  return memcmp (mac1, mac2, ETH_ALEN);
}

/* function coupleinsert()
 * this function is the higher level function for inserting data in the tables;
 * provided two "raw" piece of data (an IP and a MAC), it provides automatically
 * to insert them in the proper structures, passed through the **liste argument.
 * coupleinsert() returns -1 in case of error, or 0 if all was right.
 */
int
coupleinsert (lista_t ** liste, unsigned char newMAC[ETH_ALEN],
	      struct in_addr newIP)
{
  lista_t *MACtable, *IPtable, *newIP_l, *newMAC_l;
  struct in_addr *newIP_p;
  unsigned char *newMAC_p;
  char buffer[50];

  MACtable = liste[MAC_TYPE];
  IPtable = liste[IP_TYPE];
  mac2str (newMAC, buffer);
  err_msg (DBG_LIB, "coupleinsert(): entering");
  err_msg (DBG_LIB, "coupleinsert(): liste @%p - listaMAC @%p - listaIP @%p",
	   liste, MACtable, IPtable);
  err_msg (DBG_LOWLEV, "coupleinsert(): listaMAC:\t numreaders = %d\t"
	   " numwriters = %d", MACtable->numreaders, MACtable->numwriters);
  err_msg (DBG_LOWLEV, "coupleinsert(): listaIP:\t numreaders = %d\t"
	   " numwriters = %d", IPtable->numreaders, IPtable->numwriters);
  err_msg (DBG_LIB, "coupleinsert() trying to insert couple %s %s",
	   buffer, inet_ntoa (newIP));
  if (NULL == (newMAC_l = search (MACtable, newMAC, FINDITEM_MODE)))
    {
      if (NULL == (newMAC_p = calloc (1, ETH_ALEN * sizeof (unsigned char))))
	{
	  err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
	  return -1;
	}
      memcpy (newMAC_p, newMAC, ETH_ALEN);
    }
  else
    newMAC_p = (unsigned char *) (newMAC_l->data);
  if (NULL == (newMAC_l = insert (MACtable, newMAC_p)))
    return -1;
  err_msg (DBG_LIB, "coupleinsert(): MAC inserted");
  if (NULL == (newIP_l = search (IPtable, &newIP, FINDITEM_MODE)))
    {
      if (NULL == (newIP_p = calloc (1, sizeof (struct in_addr))))
	{
	  err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
	  return -1;
	}
      memcpy (newIP_p, &newIP, sizeof (struct in_addr));
    }
  else
    newIP_p = (struct in_addr *) (newIP_l->data);
  if (NULL == (newIP_l = insert (IPtable, newIP_p)))
    return -1;
  err_msg (DBG_LIB, "coupleinsert(): IP inserted");
/* we handle insertion in the respective sublists */
  if (NULL == sublistinsert (newIP_l, newMAC_l) ||
      NULL == sublistinsert (newMAC_l, newIP_l))
    return -1;
  relsort (IPtable, newIP_l);
  relsort (MACtable, newMAC_l);
  err_msg (DBG_LIB, "coupleinsert(): sublist handled correctly");
  err_msg (DBG_LIB, "coupleinsert(): exiting");
  return 0;
}

/* function createlists()
 * this function should be used to at first create the struct to hold data. 
 */
lista_t **
createlists (void)
{
  lista_t **liste;
  int type;

  err_msg (DBG_LIB, "createlists(): entering");
  if (NULL == (liste = calloc (2, sizeof (lista_t **))))
    {
      err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
      return NULL;
    }
  for (type = MAC_TYPE; type <= IP_TYPE; type++)
    {
      if (NULL == (liste[type] = calloc (1, sizeof (lista_t))))
	{
	  err_msg (MSG_ERR, "calloc() failed: %s", strerror (errno));
	  return NULL;
	}
      liste[type]->datatype = type;
      err_msg (DBG_LIB, "createlists(): initialising semaphores");
      semaphores_init (liste[type], LISTA_T);
    }
  err_msg (DBG_LIB, "createlists(): lists created correctly");
  err_msg (DBG_LIB, "createlists(): exiting");
  return liste;
}

/* function display_lists()
 * this function is used to print out the MAC and IP lists for debug purposes.
 */
void
display_lists (lista_t ** liste)
{
  char buffer[20];
  lista_t *cursor;
  sublist_t *subcursor;

  read_reserv (liste[IP_TYPE], LISTA_T);
  cursor = (liste[IP_TYPE])->nextnum;
  err_msg (MSG_INFO, "******** IP table summary");
  while (cursor != NULL)
    {
      err_msg (MSG_INFO, " --> IP %s found %d times, with %d relatives:",
	       inet_ntoa (*((struct in_addr *) (cursor->data))),
	       cursor->howmany, cursor->relnum);
      subcursor = cursor->relatives;
      while (subcursor != NULL)
	{
	  err_msg (MSG_INFO, "   - MAC %s, %d times.",
		   mac2str ((unsigned char *) ((subcursor->item)->data),
			    buffer), subcursor->howmany);
	  subcursor = subcursor->nextnum;
	}
      cursor = cursor->nextnum;
    }
  read_release (liste[IP_TYPE], LISTA_T);
  read_reserv (liste[MAC_TYPE], LISTA_T);
  cursor = (liste[MAC_TYPE])->nextnum;
  err_msg (MSG_INFO, "******** MAC table summary");
  while (cursor != NULL)
    {
      mac2str ((unsigned char *) (cursor->data), buffer);
      err_msg (MSG_INFO, " --> MAC %s found %d times, with %d relatives:",
	       buffer, cursor->howmany, cursor->relnum);
      subcursor = cursor->relatives;
      while (subcursor != NULL)
	{
	  err_msg (MSG_INFO, "   - IP %s, %d times.",
		   inet_ntoa (*
			      ((struct in_addr *) ((subcursor->item)->data))),
		   subcursor->howmany);
	  subcursor = subcursor->nextnum;
	}
      cursor = cursor->nextnum;
    }
  read_release (liste[MAC_TYPE], LISTA_T);
  return;
}

/* function ip2test_insert()
 * inserts a new IP address in the list pointed by head.
 */
void
ip2test_insert (ip2test_t * head, struct in_addr ip2ins)
{
  ip2test_t *cur, *backcur, *newentry;

  read_reserv (head, IPLIST_T);
  cur = head->next;
  backcur = head;
  read_release (head, IPLIST_T);
  while (cur != NULL)
    {
      read_reserv (cur, IPLIST_T);
      backcur = cur;
      cur = cur->next;
      read_release (backcur, IPLIST_T);
    }
  write_reserv (backcur, IPLIST_T);
  if (NULL == (newentry = calloc (1, sizeof (struct ip2test))))
    err_quit ("calloc() failed: %s", strerror (errno));
  semaphores_init (newentry, IPLIST_T);
  (newentry->ip).s_addr = ip2ins.s_addr;
  backcur->next = newentry;
  write_release (backcur, IPLIST_T);
  return;
}

/* function ip2test_search()
 * searches a given IP in the list pointed by head; returns a pointer to the 
 * found IP or NULL if it was not found.
 */
ip2test_t *
ip2test_search (ip2test_t * head, struct in_addr ip2search)
{
  ip2test_t *cur, *backcur;

  read_reserv (head, IPLIST_T);
  cur = head->next;
  read_release (head, IPLIST_T);
  while (cur != NULL)
    {
      read_reserv (cur, IPLIST_T);
      if (0 != ipcmp (&(cur->ip), &ip2search))
	{
	  read_release (cur, IPLIST_T);
	  break;
	}
      backcur = cur;
      cur = cur->next;
      read_release (backcur, IPLIST_T);
    }
  return cur;
}

/* function ip2test_resetlist()
 * resets a list pointed by head, clearing every IP in the list and freeing the
 * memory allocated.
 */
void
ip2test_resetlist (ip2test_t * head)
{
  ip2test_t *cur, *backcur;

  write_reserv (head, IPLIST_T);
  cur = head->next;
  head->next = NULL;
  write_release (head, IPLIST_T);
  while (cur != NULL)
    {
      write_reserv (cur, IPLIST_T);
      backcur = cur;
      cur = cur->next;
      free (backcur);
    }
  return;
}
